#!/usr/bin/env python
#
# glsh_funcs.py - Functions used by GLSH instances.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions which are used by :class:`.GLSH` instances
for rendering :class:`.Image` overlays which contain fibre orientation
distribution (FOD) spherical harmonic (SH) coefficients, in an OpenGL 2.1
compatible manner.  The functions defined in this module are intended to be
called by :class:`.GLSH` instances.


For each voxel, a sphere is drawn, with the position of each vertex on the
sphere adjusted by the SH coefficients (radii). For one draw call, the radii
for all voxels and vertices is calculated, and stored in a texture.


Different vertex/fragment shaders are used depending upon the current settings
of the :class:`.SHOpts` instance associated with the :class:`.GLSH`. If the
:attr:`.VectorOpts.colourImage` property is set, the ``glsh_volume_vert.glsl``
andf ``glvolume_frag.glsl`` shaders are used. In this case, the FODs are each
voxel are coloured according to the values in the ``colourImage``.  Otherwise,
the ``glsh_vert.glsl`` and ``glsh_frag.glsl`` shaders are used. In this case,
the vertices of each FOD are coloured according to their orientation, or to
their radius.
"""


import numpy                        as np
import numpy.linalg                 as npla

import OpenGL.GL                    as gl

import OpenGL.GL.ARB.draw_instanced as arbdi

import fsl.transform.affine         as affine
import fsleyes.gl.shaders           as shaders


def destroy(self):
    """Destroys the shader program """

    if self.shader is not None:
        self.shader.destroy()
        self.shader = None


def compileShaders(self):
    """Creates a :class:`.GLSLShader`, and attaches it to this :class:`.GLSH`
    instance as an attribute called ``shader``.
    """

    if self.shader is not None:
        self.shader.destroy()

    opts                     = self.opts
    self.useVolumeFragShader = opts.colourImage is not None

    if self.useVolumeFragShader:
        vertShader = 'glsh_volume'
        fragShader = 'glvolume'
    else:
        vertShader = 'glsh'
        fragShader = 'glsh'

    vertSrc = shaders.getVertexShader(  vertShader)
    fragSrc = shaders.getFragmentShader(fragShader)

    self.shader = shaders.GLSLShader(vertSrc, fragSrc, indexed=True)


def updateShaderState(self):
    """Updates the state of the vertex and fragment shaders. """

    shader  = self.shader
    image   = self.image
    opts    = self.opts

    if shader is None:
        return

    lightPos  = np.array([-1, -1, 4], dtype=np.float32)
    lightPos /= np.sqrt(np.sum(lightPos ** 2))

    shape = image.shape[:3]
    xFlip = opts.orientFlip

    if   opts.colourMode == 'direction': colourMode = 0
    elif opts.colourMode == 'radius':    colourMode = 1

    modLow,  modHigh  = self.getModulateRange()
    clipLow, clipHigh = self.getClippingRange()
    modMode           = {'brightness' : 0,
                         'alpha'      : 1}[opts.modulateMode]

    clipXform   = self.getAuxTextureXform('clip')
    colourXform = self.getAuxTextureXform('colour')
    modXform    = self.getAuxTextureXform('modulate')

    shader.load()

    changed  = False
    changed |= shader.set('xFlip',       xFlip)
    changed |= shader.set('imageShape',  shape)
    changed |= shader.set('lighting',    opts.lighting)
    changed |= shader.set('lightPos',    lightPos)
    changed |= shader.set('nVertices',   self.vertices.shape[0])
    changed |= shader.set('sizeScaling', opts.size / 100.0)
    changed |= shader.set('radTexture',  4)

    if self.useVolumeFragShader:

        voxValXform     = self.colourTexture.voxValXform
        invVoxValXform  = self.colourTexture.invVoxValXform
        texZero         = 0.0 * invVoxValXform[0, 0] + invVoxValXform[0, 3]
        img2CmapXform   = affine.concat(
            self.cmapTexture.getCoordinateTransform(),
            voxValXform)

        changed |= shader.set('clipTexture',      1)
        changed |= shader.set('imageTexture',     2)
        changed |= shader.set('colourTexture',    3)
        changed |= shader.set('negColourTexture', 3)
        changed |= shader.set('img2CmapXform',    img2CmapXform)
        changed |= shader.set('imageIsClip',      False)
        changed |= shader.set('useNegCmap',       False)
        changed |= shader.set('useSpline',        False)
        changed |= shader.set('clipLow',          clipLow)
        changed |= shader.set('clipHigh',         clipHigh)
        changed |= shader.set('texZero',          texZero)
        changed |= shader.set('invertClip',       False)
        changed |= shader.set('colourCoordXform', colourXform)
        changed |= shader.set('clipCoordXform',   clipXform)

    else:

        cmapXform            = self.cmapTexture.getCoordinateTransform()
        colours, colourXform = self.getVectorColours()

        changed |= shader.set('modulateTexture',  0)
        changed |= shader.set('clipTexture',      1)
        changed |= shader.set('cmapTexture',      3)
        changed |= shader.set('clipLow',          clipLow)
        changed |= shader.set('clipHigh',         clipHigh)
        changed |= shader.set('modLow',           modLow)
        changed |= shader.set('modHigh',          modHigh)
        changed |= shader.set('modulateMode',     modMode)
        changed |= shader.set('colourMode',       colourMode)
        changed |= shader.set('xColour',          colours[0])
        changed |= shader.set('yColour',          colours[1])
        changed |= shader.set('zColour',          colours[2])
        changed |= shader.set('colourXform',      colourXform)
        changed |= shader.set('cmapXform',        cmapXform)
        changed |= shader.set('clipCoordXform',   clipXform)
        changed |= shader.set('modCoordXform',    modXform)

    shader.setAtt('vertex',   self.vertices)
    shader.setAtt('vertexID', self.vertIdxs)
    shader.setIndices(        self.indices)

    shader.unload()

    return changed


def preDraw(self, xform=None, bbox=None):
    """Called by :meth:`.GLSH.preDraw`. Loads the shader program, and updates
    some shader attributes.
    """
    shader = self.shader

    shader.load()

    # Calculate a transformation matrix for
    # normal vectors - T(I(MV matrix))
    # We transpose mvMat because OpenGL is column-major
    mvMat        = gl.glGetFloatv(gl.GL_MODELVIEW_MATRIX)[:3, :3].T
    v2dMat       = self.opts.getTransform('voxel', 'display')[:3, :3]

    normalMatrix = affine.concat(mvMat, v2dMat)
    normalMatrix = npla.inv(normalMatrix).T

    shader.set('normalMatrix', normalMatrix)

    gl.glEnable(gl.GL_CULL_FACE)
    gl.glClear(gl.GL_DEPTH_BUFFER_BIT)
    gl.glEnable(gl.GL_DEPTH_TEST)
    gl.glCullFace(gl.GL_BACK)


def draw2D(self, zpos, axes, xform=None, bbox=None):
    """Called by :meth:`.GLSH.draw2D`. Draws the scene. """

    opts   = self.opts
    shader = self.shader
    v2dMat = opts.getTransform('voxel',   'display')

    if xform is None: xform = v2dMat
    else:             xform = affine.concat(v2dMat, xform)

    voxels              = self.generateVoxelCoordinates2D(zpos, axes, bbox)
    voxels, radTexShape = self.updateRadTexture(voxels)

    if len(voxels) == 0:
        return

    voxIdxs = np.arange(voxels.shape[0], dtype=np.float32)

    shader.setAtt('voxel',           voxels,  divisor=1)
    shader.setAtt('voxelID',         voxIdxs, divisor=1)
    shader.set(   'voxToDisplayMat', xform)
    shader.set(   'radTexShape',     radTexShape)
    shader.set(   'radXform',        self.radTexture.voxValXform)

    shader.loadAtts()

    arbdi.glDrawElementsInstancedARB(
        gl.GL_TRIANGLES, self.nVertices, gl.GL_UNSIGNED_INT, None, len(voxels))


def draw3D(self, xform=None, bbox=None):
    pass


def postDraw(self, xform=None, bbox=None):
    """Called by :meth:`.GLSH.draw`. Cleans up the shader program and GL
    state.
    """

    self.shader.unloadAtts()
    self.shader.unload()
    gl.glDisable(gl.GL_CULL_FACE)
    gl.glDisable(gl.GL_DEPTH_TEST)



#!/usr/bin/env python
#
# The profiles module contains logic for mouse-keyboard interaction.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The :mod:`profiles` package contains logic for mouse/keyboard interaction
with :class:`.ViewPanel` panels.


This logic is encapsulated in four classes:


 - The :class:`Profile` class is intended to be subclassed. A :class:`Profile`
   instance contains the mouse/keyboard event handlers for a particular type
   of ``ViewPanel`` to allow the user to interact with the view in a
   particular way. For example, the :class:`.OrthoViewProfile` class allows
   the user to navigate through the display space in an :class:`.OrthoPanel`
   canvas, whereas the :class:`.OrthoEditProfile` class contains interaction
   logic for selecting and editing :class:`.Image` voxels in an ``OrthoPanel``.


 - The :class:`.CanvasPanelEventManager` manages ``wx`` GUI events on the
   :class:`.SliceCanvas` instances contained in :class:`.CanvasPanel` views.


 - The :class:`.PlotPanelEventManager` manages ``matplotlib`` GUI events on
   the ``matplotlib Canvas`` instances contained in :class:`.PlotPanel`
   views.


 - The :class:`ProfileManager` class is used by ``ViewPanel`` instances to
   create and change the ``Profile`` instance currently in use.


The ``profiles`` package is also home to the :mod:`.shortcuts` module, which
defines global *FSLeyes* keyboard shortcuts.
"""


import logging
import inspect
import collections

import wx

import matplotlib.backend_bases as mplbackend

import fsl.utils.notifier   as notifier
import fsleyes_widgets      as fwidgets
import fsleyes_props        as props
import fsleyes.actions      as actions


log = logging.getLogger(__name__)


class ProfileManager(notifier.Notifier):
    """Manages creation/registration/de-registration of :class:`Profile`
    instances for a :class:`.ViewPanel` instance.

    A :class:`ProfileManager` instance is created and used by every
    :class:`.ViewPanel` instance. The ``ProfileManager`` manages a stack of
    :class:`.Profile` instances - profiles can be activated and deactivated
    via the :meth:`activateProfile` and :meth:`deactivateProfile` methods.

    The ``ProfileManager`` uses the :class:`.Notifier` interface to notify any
    interested listeners when the current profile is changed, using the topic
    name ``'profile'``. Registered listeners are passed a tuple containing the
    types (:class:`.Profile` sub-classes) of the de-registered and newly
    registered profiles.
    """


    def __init__(self, viewPanel, overlayList, displayCtx, maxprofiles):
        """Create a :class:`ProfileManager`.

        :arg viewPanel:      The :class:`.ViewPanel` instance which this
                             :class:`ProfileManager` is to manage.

        :arg overlayList:    The :class:`.OverlayList` instance containing the
                             overlays that are being displayed.

        :arg displayCtx:     The :class:`.DisplayContext` instance which
                             defines how overlays are being displayed.

        :arg maxprofiles:    Maximum numbe of profiles which can be in the
                             profile stack.
        """

        self.__viewPanel   = viewPanel
        self.__viewCls     = viewPanel.__class__
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx
        self.__maxprofiles = maxprofiles

        # a stack of profiles - [0] is the
        # default, and [-1] is currently active
        self.__profileStack = []


    def destroy(self):
        """This method must be called by the owning :class:`.ViewPanel` when
        it is about to be destroyed (or when it no longer needs a
        ``ProfileManager``).

        This method destroys all class:`Profile` instances, and clears some
        internal object references to avoid memory leaks.
        """

        if self.numProfiles() > 0:
            self.__profileStack[-1].deregister()

        for prof in self.__profileStack:
            prof.destroy()

        self.__profileStack = []
        self.__viewPanel    = None
        self.__overlayList  = None
        self.__displayCtx   = None


    def numProfiles(self):
        """Returns the number of profiles that currently exist. """
        return len(self.__profileStack)


    def getCurrentProfile(self):
        """Returns the :class:`Profile` instance currently in use."""

        if self.numProfiles() > 0:
            return self.__profileStack[-1]
        else:
            return None


    def deactivateProfile(self, notify=True):
        """Deactivates and destroys the current profile, and re-activates
        the previous one.

        :arg notify: If ``True`` (the default), a notification is emitted
                     via the :class:`.Notifier` interface.
        :returns:    A reference to the re-activated profile.
        """

        oldprof = self.__profileStack.pop()

        log.debug('Deregistering %s profile from view %s',
                  type(oldprof).__name__, self.__viewCls.__name__)

        oldprof.deregister()
        oldprof.destroy()

        if self.numProfiles() > 0:
            self.__profileStack[-1].register()

        newprof = self.getCurrentProfile()

        if notify:
            if newprof is None: value = (type(oldprof), None)
            else:               value = (type(oldprof), type(newprof))
            self.notify(topic='profile', value=value)

        return newprof


    def activateProfile(self, profileCls, notify=True):
        """Deregisters the current :class:`Profile` instance, and creates and
        registers a new instance of type ``profileCls``.

        :arg notify: If ``True`` (the default), a notification is emitted
                     via the :class:`.Notifier` interface.
        :returns:    A reference to the newly registered profile.
        """

        if self.numProfiles() == self.__maxprofiles:
            raise RuntimeError('Cannot create any more profiles!')

        if self.numProfiles() > 0:
            self.__profileStack[-1].deregister()

        log.debug('Creating and registering profile %s with view %s',
                  profileCls.__name__, type(self.__viewPanel).__name__)

        oldprof = self.getCurrentProfile()
        newprof = profileCls(self.__viewPanel,
                             self.__overlayList,
                             self.__displayCtx)

        self.__profileStack.append(newprof)
        self.__profileStack[-1].register()

        if notify:
            if oldprof is None: value = (None,          type(newprof))
            else:               value = (type(oldprof), type(newprof))
            self.notify(topic='profile', value=value)

        return newprof


class Profile(props.SyncableHasProperties, actions.ActionProvider):
    """A :class:`Profile` class implements keyboard/mouse interaction behaviour
    for a :class:`.ViewPanel` instance.


    Subclasses should specify at least one *mode* of operation, which defines
    a sort of sub-profile. The current mode can be changed with the
    :attr:`mode` property.


    Subclasses must also override the :meth:`getEventTargets` method, to
    return the :mod:`wx` objects that are to be the targets for mouse/keyboard
    interaction.


    The ``Profile`` class currently only supports :class:`.CanvasPanel` and
    :class:`.PlotPanel` views. ``Profile`` instances use a
    :class:`.CanvasPanelEventManager` instance to manage GUI events on
    :class:`.CanvasPanel` instances, or a:class:`.PlotPanelEventManager`
    to manage GUI events on ``matplotlib Canvas`` objects.


    **Receiving events**


    In order to receive mouse or keyboard events, subclasses simply need to
    implement methods which handle the events of interest for the relevant
    mode, and name them appropriately. The name of a method handler must be
    of the form::


        _[modeName]Mode[eventName]


    where ``modeName`` is an identifier for the profile mode (see the
    :meth:`__init__` method), and ``eventName`` is one of the following:

      - ``LeftMouseMove``
      - ``LeftMouseDown``
      - ``LeftMouseDrag``
      - ``LeftMouseUp``
      - ``RightMouseMove``
      - ``RightMouseDown``
      - ``RightMouseDrag``
      - ``RightMouseUp``
      - ``MiddleMouseMove``
      - ``MiddleMouseDown``
      - ``MiddleMouseDrag``
      - ``MiddleMouseUp``
      - ``MouseWheel``
      - ``Char``

    Profiles for :class:`.CanvasPanel` views may also implement handlers
    for these events:

      - ``MouseEnter``
      - ``MouseLeave``

    And :class:`.PlotPanel` views may implement handlers for these events:

      - ``LeftMouseArtistPick``
      - ``RightMouseArtistPick``


    .. note:: The ``MouseEnter`` and ``MouseLeave`` events are not supported
              on :class:`.PlotPanel` views due to bugs in ``matplotlib``.


    For example, if a particular profile has defined a mode called ``nav``,
    and is interested in left clicks, the profile class must provide a method
    called ``_navModeLeftMouseDown``. Then, whenever the profile is in the
    ``nav`` mode, this method will be called on left mouse clicks.


    **Handler methods**


    The parameters that are passed to these methods differs slightly depending
    on the type of event:

     - All mouse events, with the exception of ``MouseWheel`` must have
       the following signature::

        def _[modeName]Mode[eventName](ev, canvas, mouseLoc, canvasLoc)

       where:

         - ``ev`` is the ``wx.Event`` object
         - ``canvas`` is the source canvas,
         - ``mouseLoc`` is the ``(x, y)`` mouse coordinates,
         - ``canvasLoc`` is the coordinates in the display/data coordinate
           system.

     - The ``MouseWheel`` handler must have the following signature::

        def _[modeName]ModeMouseWheel(ev, canvas, wheel, mouseLoc, canvasLoc)

       where ``wheel`` is a positive or negative number indicating how much
       the mouse wheel was moved.

     - ``Char`` events must have the following signature::

        def _[modeName]ModeChar(ev, canvas, key)

       where ``key`` is the key code of the key that was pressed.

     - Pick event handlers (only on :class:`.PlotPanel` views) must have
       the following signature::

        def _[modeName]Mode[eventType](ev, canvas, artist, mouseLoc, canvasLoc)

       where ``artist`` is the ``matplotlib`` artist that was picked.

    All handler methods should return ``True`` to indicate that the event was
    handled, or ``False`` if the event was not handled. This is particularly
    important for ``Char`` event handlers - we don't want ``Profile``
    sub-classes to be eating global keyboard shortcuts. A return value of
    ``None`` is interpreted as ``True``. If a handler returns ``False``, and
    a fallback handler is defined (see below), then that fallback handler will
    be called.


    **Extra handlers**


    Additional handlers can be registered for any event type via the
    :meth:`registerHandler` method. These handlers do not have to be methods
    of the ``Profile`` sub-class, and will be called for every occurrence of
    the event, regardless of the current mode. These handlers will be called
    after the standard handler method.


    When an extra handler is no longer needed, it must be removed via the
    :meth:`deregisterHandler` method.


    **Pre- and post- methods**


    A couple of other methods may be defined which, if they are present, will
    be called on all handled events:

     - ``_preEvent``
     - ``_postEvent``

    The ``_preEvent`` method will get called just before an event is passed
    to the handler. Likewise, the ``_postEvent`` method will get called
    just after the handler has been called. If no handlers for a particular
    event are defined, neither of these methods will be called.


    **Temporary, alternate and fallback handlers**


    The :meth:`tempModes` function may be overridden to define a keyboard
    modifier which may be used to temporarily redirect mouse/keyboard events
    to the handlers for a different mode. For example, if while in ``nav``
    mode, you would like the user to be able to switch to ``zoom`` mode with
    the control key, you can add a temporary mode map in the dictionary
    returned by ``tempModes``. Additional temporary modes can be added via the
    :meth:`addTempMode` method.

    The :meth:`altHandlers` function may be overridden to allow the re-use of
    event handlers that have been defined for one mode in another mode. For
    example, if you would like right clicks in ``zoom`` mode to behave like
    left clicks in ``nav`` mode, you can return such a mapping from the
    ``altHandler`` function. Additional alternate handlers can be added
    via the :meth:`addAltHandler` method.

    The :meth:`fallbackHandlers` function may be overridden to define fallback
    handlers - if the default handler for a specific mode/event type returns a
    value of ``False``, the event will be forwarded to the fallback handler
    instead. Additional fallback handlers can be added via the
    :meth:`addFallbackHandler` method.


    **Actions and attributes**


    As the ``Profile`` class derives from the :class:`.ActionProvider`
    class, ``Profile`` subclasses may define properties and actions for
    the user to configure the profile behaviour, and/or to perform any
    relevant actions.


    The following instance attributes are present on a ``Profile`` instance,
    intended to be accessed by sub-classes:

    =============== =======================================================
    ``viewPanel``   The :class:`ViewPanel` which is using this ``Profile``.
    ``overlayList`` A :class:`.OverlayList` instance.
    ``displayCtx``  A :class:`.DisplayContext` instance.
    ``name``        A unique name for this ``Profile`` instance.
    =============== =======================================================
    """


    mode = props.Choice()
    """The current profile mode - by default this is empty, but subclasses
    may specify the choice options in the :class:`__init__` method.
    """


    @staticmethod
    def supportedView():
        """Returns the :class:`.ViewPanel` type that is supported by this
        ``Profile``. Must be implemented by sub-classes.
        """
        raise NotImplementedError()


    @staticmethod
    def tempModes():
        """May be overridden by sub-classes.  Should return a dictionary
        defining temporary modes which, when in a given mode, can be accessed
        with a keyboard modifer (e.g. Control, Shift, etc). For example, a
        temporary mode map of::

            ('view', wx.WXK_SHIFT) : 'zoom'

        states that when the ``Profile`` is in ``'view'`` mode, and the shift
        key is held down, the ``Profile`` should temporarily switch to
        ``'zoom'`` mode.

        For multi-key combinations, the modifier key IDs must be provided as a
        tuple, in alphabetical order. For example, to specify shift+ctrl, the
        tuple must be (wx.WXK_CTRL, wx.WXK_SHIFT)

        Important: Any temporary modes which use CTRL, ALT, or CTRL+ALT must
        not handle character events, as these modifiers are reserved for
        global shortcuts.

        Temporary modes honour the ``Profile`` class hierarchy, so if you
        sub-class an existing ``Profile`` class, your class will inherit all
        of the temporary modes defined on the base class.
        """
        return None


    @staticmethod
    def altHandlers():
        """May be overridden by sub-classes. Should return a dictionary
        defining alternate handlers for a given mode and event type. Entries
        in this dictionary allow a :class:`.Profile` sub-class to define a
        handler for a single mode and event type, but to re-use that handler
        for other modes and event types. For example, the following alternate
        handler mapping::

            ('zoom', 'MiddleMouseDrag') : ('pan',  'LeftMouseDrag')

        states that when the ``Profile`` is in ``'zoom'`` mode, and a
        ``MiddleMouseDrag`` event occurs, the ``LeftMouseDrag`` handler for
        the ``'pan'`` mode should be called.

        Alternate handlers honour the ``Profile`` class hierarchy, so if you
        sub-class an existing ``Profile`` class, your class will inherit all
        of the alternate handlers defined on the base class.

        .. note:: Event bindings defined by ``altHandlers`` take precdence
                  over the event bindings defined in the :class:`.Profile`
                  sub-class. So you can use the ``altHandlers`` to override
                  the default behaviour of a ``Profile``.
        """
        return None


    @staticmethod
    def fallbackHandlers():
        """May be overridden by sub-classes. Should return a dictionary
        defining handlers for a given mode and event type which will be called
        if the handler for that mode/event type returns a value of ``False``,
        indicating that it has not been handled. For example, the
        following fallback handler mapping::

            (('pick', 'LeftMouseDown'), ('nav', 'LeftMouseDown')),

        states that when the profile is in ``'pick'`` mode, and the
        ``LeftMouseDown`` handler for ``'pick'`` mode returns ``False``, the
        ``LeftMouseDown`` handler for ``'nav'`` mode will be called.

        Fallback handlers honour the ``Profile`` class hierarchy, so if you
        sub-class an existing ``Profile`` class, your class will inherit all
        of the fallback handlers defined on the base class.
        """
        return None


    def __init__(self,
                 viewPanel,
                 overlayList,
                 displayCtx,
                 modes=None):
        """Create a ``Profile`` instance.

        :arg viewPanel:   The :class:`.ViewPanel` instance for which this
                          ``Profile`` instance defines mouse/keyboard
                          interaction behaviour.

        :arg overlayList: The :class:`.OverlayList` instance which contains
                          the list of overlays being displayed.

        :arg displayCtx:  The :class:`.DisplayContext` instance which defines
                          how the overlays are to be displayed.

        :arg modes:       A sequence of strings, containing the mode
                          identifiers for this profile. These are added as
                          options on the :attr:`mode` property.
        """

        if not isinstance(viewPanel, self.supportedView()):
            raise ValueError('Unsupported view: {}'.format(
                type(viewPanel).__name__))

        actions.ActionProvider     .__init__(self, overlayList, displayCtx)
        props.SyncableHasProperties.__init__(self)

        self.__viewPanel = viewPanel
        self.__name      = '{}_{}'.format(self.__class__.__name__, id(self))

        import fsleyes.views.canvaspanel as canvaspanel
        import fsleyes.views.plotpanel   as plotpanel

        if isinstance(viewPanel, canvaspanel.CanvasPanel):
            self.__evtManager = CanvasPanelEventManager(self)

        elif isinstance(viewPanel, plotpanel.PlotPanel):
            self.__evtManager = PlotPanelEventManager(self)

        else:
            raise ValueError('Unrecognised view panel type: {}'.format(
                type(viewPanel).__name__))

        # Maps which define temporary modes and
        # alternate/fallback handlers when keyboard
        # modifiers are used, when a handler for
        # a particular event is not defined, or
        # when a handler indicates that the event
        # has not been handled.
        self.__tempModeMap        = {}
        self.__altHandlerMap      = {}
        self.__fallbackHandlerMap = {}

        # Extra handlers - for each event type,
        # a dictionary of { name : handler }
        # mappings which will be called after
        # the profile handler has been called
        # for a given event.
        self.__extraHandlers = collections.defaultdict(collections.OrderedDict)

        # of mouse/canvas event locations
        self.__lastCanvas       = None
        self.__lastMousePos     = None
        self.__lastCanvasPos    = None
        self.__lastMouseUpPos   = None
        self.__lastCanvasUpPos  = None
        self.__mouseDownPos     = None
        self.__canvasDownPos    = None

        # we keep track of the mode we
        # were in on mosue down events,
        # so the correct mode is called
        # on subsequent drag/up events.
        self.__mouseDownMode    = None

        # This field is used to keep
        # track of the last event for
        # which a handler was called.
        # After the first event, it
        # will be a tuple of strings
        # containing the (mode, event),
        # e.g. ('nav', 'LeftMouseMove').
        # This is set in the __getHandler
        # method.
        #
        # The lastMouseUpHandler field
        # is set in __onMouseUp, to
        # keep track of the last mouse
        # handler
        self.__lastHandler        = (None, None)
        self.__lastMouseUpHandler = (None, None)

        # Pre/post event handlers
        self.__preEventHandler  = getattr(self, '_preEvent',  None)
        self.__postEventHandler = getattr(self, '_postEvent', None)

        # Add all of the provided modes
        # as options to the mode property
        if modes is None:
            modes = []

        modeProp = self.getProp('mode')

        for mode in modes:
            modeProp.addChoice(mode, instance=self)

        if len(modes) > 0:
            self.mode = modes[0]

        # Configure temporary modes and alternate
        # event handlers. We reverse the mro, so
        # that the modes/handlers defined on this
        # class take precedence.
        mro = list(reversed(inspect.getmro(self.__class__)))
        mro = mro[mro.index(Profile):]
        for cls in mro:

            tempModes   = cls.tempModes()        or {}
            altHandlers = cls.altHandlers()      or {}
            fbHandlers  = cls.fallbackHandlers() or {}

            for (mode, keymod), tempMode in tempModes.items():
                self.addTempMode(mode, keymod, tempMode)

            for (mode, handler), (altMode, altHandler) in altHandlers.items():
                self.addAltHandler(mode, handler, altMode, altHandler)

            for (mode, handler), (fbMode, fbHandler) in fbHandlers.items():
                self.addFallbackHandler(mode, handler, fbMode, fbHandler)

        # The __onEvent method delegates all
        # events based on this dictionary
        self.__eventMap = {
            wx.EVT_LEFT_DOWN.typeId    : self.__onMouseDown,
            wx.EVT_MIDDLE_DOWN.typeId  : self.__onMouseDown,
            wx.EVT_RIGHT_DOWN.typeId   : self.__onMouseDown,
            wx.EVT_LEFT_UP.typeId      : self.__onMouseUp,
            wx.EVT_MIDDLE_UP.typeId    : self.__onMouseUp,
            wx.EVT_RIGHT_UP.typeId     : self.__onMouseUp,
            wx.EVT_MOTION.typeId       : self.__onMouseMove,
            wx.EVT_MOUSEWHEEL.typeId   : self.__onMouseWheel,
            wx.EVT_ENTER_WINDOW.typeId : self.__onMouseEnter,
            wx.EVT_LEAVE_WINDOW.typeId : self.__onMouseLeave,
            wx.EVT_CHAR.typeId         : self.__onChar,
        }

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message. """
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    def destroy(self):
        """This method must be called when this ``Profile`` is no longer
        needed - it is typically called by a :class:`ProfileManager`.

        Clears references to the display context, view panel, and overlay
        list, and calls :meth:`.ActionProvider.destroy`.
        """
        actions.ActionProvider.destroy(self)
        self.__lastCanvas    = None
        self.__viewPanel     = None
        self.__extraHandlers = None


    @property
    def name(self):
        """Returns the name of this ``Profile``. """
        return self.__name


    @property
    def viewPanel(self):
        """Returns the :class:`.ViewPanel` associated with this ``Profile``.
        """
        return self.__viewPanel


    def getEventTargets(self):
        """Must be overridden by subclasses, to return a sequence of
        :mod:`wx` objects that are the targets of mouse/keyboard interaction.

        .. note:: It is currently assumed that all of the objects in the
                  sequence derive from the :class:`.SliceCanvas` class.
        """
        raise NotImplementedError('Profile subclasses must implement '
                                  'the getEventTargets method')


    def getMouseDownLocation(self):
        """If the mouse is currently down, returns a 2-tuple containing the
        x/y mouse coordinates, and the corresponding 3D display space
        coordinates, of the mouse down event. Otherwise, returns
        ``(None, None)``.
        """
        return self.__mouseDownPos, self.__canvasDownPos


    def getLastMouseLocation(self):
        """Returns a 2-tuple containing the most recent x/y mouse coordinates,
        and the corresponding 3D display space coordinates.
        """
        return self.__lastMousePos, self.__lastCanvasPos


    def getLastMouseUpLocation(self):
        """Returns a 2-tuple containing the most recent x/y mouse up event
        coordinates, and the corresponding 3D display space coordinates.
        """
        return self.__lastMouseUpPos, self.__lastCanvasUpPos


    def getLastMouseUpHandler(self):
        """Returns a tuple of two strings specifying the ``(mode, eventType)``
        of the most recent mouse up event that was handled. If no events have
        been handled, returns ``(None, None)``.
        """
        return self.__lastMouseUpHandler


    def getLastHandler(self):
        """Returns a tuple of two strings specifying the ``(mode, eventType)``
        of the most recent event that was handled. If no events have been
        handled, returns ``(None, None)``.
        """
        return self.__lastHandler


    def getLastCanvas(self):
        """Returns a reference to the canvas which most recently generated
        a mouse down or up event.
        """
        return self.__lastCanvas


    def getMplEvent(self):
        """If this ``Profile`` object is associated with a :class:`.PlotPanel`,
        this method will return the last ``matplotlib`` event that was
        generated. Otherwise, this method returns ``None``.

        This method can be called from within an event handler to retrieve
        the current ``matplotlib`` event. See the
        :meth:`PlotPanelEventManager.getMplEvent` method.
        """
        if isinstance(self.__evtManager, PlotPanelEventManager):
            return self.__evtManager.getMplEvent()
        else:
            return None


    def addTempMode(self, mode, modifier, tempMode):
        """Add a temporary mode to this ``Profile``, in addition to those
        defined by the :meth:`tempModes` function.

        :arg mode:     The mode to change from.

        :arg modifier: A keyboard modifier which will temporarily
                       change the mode from ``mode`` to ``tempMode``.

        :arg tempMode: The temporary mode which the ``modifier`` key will
                       change into.
        """
        self.__tempModeMap[mode, modifier] = tempMode


    def addAltHandler(self, mode, event, altMode, altEvent):
        """Add an alternate handler to this ``Profile``, in addition to
        those already defined by the :meth:`altHandlers` function.

        :arg mode:     The source mode.

        :arg event:    Name of the event to handle (e.g. ``LeftMouseDown``).

        :arg altMode:  The mode for which the handler is defined.

        :arg altEvent: The event name for which the handler is defined.
        """
        self.__altHandlerMap[mode, event] = (altMode, altEvent)


    def addFallbackHandler(self, mode, event, fbMode, fbEvent):
        """Add a fallback handler to this ``Profile``, in addition to
        those already defined by the :meth:`fallbackHandlers` function.

        :arg mode:    The source mode.

        :arg event:   Name of the event to handle (e.g. ``LeftMouseDown``).

        :arg fbMode:  The mode for which the handler is defined.

        :arg fbEvent: The event name for which the handler is defined.
        """
        self.__fallbackHandlerMap[mode, event] = (fbMode, fbEvent)


    def register(self):
        """This method must be called to register this ``Profile``
        instance as the target for mouse/keyboard events. This method
        is called by the :class:`ProfileManager`.

        Subclasses may override this method to performa any initialisation,
        but must make sure to call this implementation.
        """
        self.__evtManager.register()


    def deregister(self):
        """This method de-registers this :class:`Profile` instance from
        receiving mouse/keybouard events. This method is called by the
        :class:`ProfileManager`.


        Subclasses may override this method to performa any initialisation,
        but must make sure to call this implementation.
        """
        self.__evtManager.deregister()


    def registerHandler(self, event, name, handler):
        """Add an extra handler for the specified event.

        When the event occurs, The ``handler`` function will be called after
        the default handler, provided by the  ``Profie`` sub-class, is called.

        :arg event:   The event type (e.g. ``LeftMouseDown``).

        :arg name:    A unique name for the handler. A ``KeyError`` will be
                      raised if a handler with ``name`` is already registered.

        :arg handler: Function to call when the event occurs. See the class
                      documentation for details on the required signature.
        """

        if name in self.__extraHandlers[event]:
            raise KeyError('A handler with name "{}" is '
                           'already registered'.format(name))

        self.__extraHandlers[event][name] = handler


    def deregisterHandler(self, event, name):
        """Remove an extra handler from the specified event, that was previously
        added via :meth:`registerHandler`

        :arg event:   The event type (e.g. ``LeftMouseDown``).

        :arg name:    A unique name for the handler. A ``KeyError`` will be
                      raised if a handler with ``name`` is already registered.
        """
        if self.__extraHandlers is not None:
            self.__extraHandlers[event].pop(name)


    def handleEvent(self, ev):
        """Called by the event manager when any event occurs on any of
        the :class:`.ViewPanel` targets. Delegates the event to one of
        the handler functions.

        :arg ev: The ``wx.Event`` that occurred.
        """

        evType  = ev.GetEventType()
        source  = ev.GetEventObject()
        handler = self.__eventMap.get(evType, None)

        if source not in self.getEventTargets(): return
        if handler is None:                      return

        if evType in (wx.EVT_LEFT_DOWN  .typeId,
                      wx.EVT_MIDDLE_DOWN.typeId,
                      wx.EVT_RIGHT_DOWN .typeId,
                      wx.EVT_LEFT_UP    .typeId,
                      wx.EVT_MIDDLE_UP  .typeId,
                      wx.EVT_RIGHT_UP   .typeId):
            self.__lastCanvas = source
        handler(ev)


    def handlePickEvent(self, ev):
        """Called by the :class:`PlotPanelEventManager` when a ``matplotlib``
        ``pick_event`` occurs.
        """

        self.__onPick(ev)


    def __getTempMode(self, ev):
        """Checks the temporary mode map to see if a temporary mode should
        be applied. Returns the mode identifier, or ``None`` if no temporary
        mode is applicable.
        """

        mode  = self.mode
        alt   = ev.AltDown()
        ctrl  = ev.ControlDown()
        shift = ev.ShiftDown()

        # Figure out the dictionary key to use,
        # based on the modifier keys that are down
        keys  = {
            (False, False, False) :  None,
            (False, False, True)  :  wx.WXK_SHIFT,
            (False, True,  False) :  wx.WXK_CONTROL,
            (False, True,  True)  : (wx.WXK_CONTROL, wx.WXK_SHIFT),
            (True,  False, False) :  wx.WXK_ALT,
            (True,  False, True)  : (wx.WXK_ALT, wx.WXK_SHIFT),
            (True,  True,  False) : (wx.WXK_ALT, wx.WXK_CONTROL),
            (True,  True,  True)  : (wx.WXK_ALT, wx.WXK_CONTROL, wx.WXK_SHIFT)
        }

        return self.__tempModeMap.get((mode, keys[alt, ctrl, shift]), None)


    def __getMouseLocation(self, ev):
        """Returns two tuples; the first contains the x/y coordinates of the
        given :class:`wx.MouseEvent`, and the second contains the
        corresponding x/y/z display space coordinates (for
        :class:`.CanvasPanel` views), or x/y data coordinates (for
        :class:`.PlotPanel` views).

        See the :meth:`CanvasPanelEventManager.getMouseLocation` and
        :meth:`PlotPanelEventManager.getMouseLocation` methods.
        """
        return self.__evtManager.getMouseLocation(ev)


    def __getMouseButton(self, ev):
        """Returns a string describing the mouse button associated with the
        given :class:`wx.MouseEvent`.
        """

        btn = ev.GetButton()
        if   btn == wx.MOUSE_BTN_LEFT:   return 'Left'
        elif btn == wx.MOUSE_BTN_RIGHT:  return 'Right'
        elif btn == wx.MOUSE_BTN_MIDDLE: return 'Middle'
        elif ev.LeftIsDown():            return 'Left'
        elif ev.RightIsDown():           return 'Right'
        elif ev.MiddleIsDown():          return 'Middle'
        else:                            return  None


    def __getMode(self, ev):
        """Returns the current profile mode - either the value of
        :attr:`mode`, or a temporary mode if one is active.
        """

        # Is a temporary mode active?
        tempMode = self.__getTempMode(ev)

        if tempMode is None: return self.mode
        else:                return tempMode


    def __getHandler(self,
                     ev,
                     evType,
                     mode=None,
                     origEvType=None,
                     direct=False):
        """Returns a function which will handle the given
        :class:`wx.MouseEvent` or :class:`wx.KeyEvent` (the ``ev`` argument),
        or ``None`` if no handlers are found.

        If an alternate handler for the mode/event has been specified, it is
        returned.

        :arg ev:         The event object

        :arg evType:     The event type (e.g. ``'LeftMouseDown'``)

        :arg mode:       Override the default mode with this one. If not
                         provided, the handler for the current mode (or
                         temporary mode, if one is active) will be used.

        :arg origEvType: If the ``evType`` is not the actual event that
                         occurred (e.g. this method has been called to look
                         up an alternate or fallback handler), the original
                         event type must be passed in here.

        :arg direct:     If ``False`` (the default), the returned function will
                         call the standard event handler (a method of the
                         ``Profile`` sub-class), its fallback handler if it
                         returns ``False`` and a fallback has been specfiied,
                         any extra handlers that have been registered for the
                         event type, and will also call the pre- and post-
                         event methods. Otherwise, the returned function will
                         just be the sub-class handler method for the
                         specified for ``evType/ ``mode``.
        """

        if origEvType is None:
            origEvType = evType

        if mode is None:
            mode = self.__getMode(ev)

        # Lookup any alternate/fallback
        # handlers for the event
        alt      = self.__altHandlerMap     .get((mode, evType), None)
        fallback = self.__fallbackHandlerMap.get((mode, evType), None)

        # Is an alternate handler active?
        # Alternate handlers take precedence
        # over default handlers.
        if alt is not None:
            altMode, altEvType = alt
            return self.__getHandler(ev,
                                     altEvType,
                                     mode=altMode,
                                     origEvType=evType)

        # A fallback handler has
        # been specified for this
        # event - get a direct ref
        # to the fallback function
        if fallback is not None:
            fbMode, fbEvType = fallback
            fallback = self.__getHandler(ev,
                                         fbEvType,
                                         mode=fbMode,
                                         origEvType=evType,
                                         direct=True)

        # Search for a default method
        # which can handle the specified
        # mode/evtype.
        if mode is not None:
            handlerName = '_{}Mode{}'.format(mode, evType)
        else:
            handlerName = '_{}{}'.format(evType[0].lower(), evType[1:])

        defHandler = getattr(self, handlerName, None)

        # If direct=True, we just
        # return the handler method,
        # even it there isn't one
        # defined.
        if direct:
            return defHandler

        # Otherwise we return a wrapper
        # which calls the pre- and post-
        # methods, and any extra handlers
        # that have been registered,
        handlers = []

        # Insert a placeholder for the
        # default handler, because we
        # need to check its return value.
        if defHandler is not None:
            handlers.append('defHandler')

        handlers.extend(self.__extraHandlers[origEvType].values())

        def handlerWrapper(*args, **kwargs):

            retval = None

            if self.__preEventHandler is not None:
                self.__preEventHandler(mode, evType)

            for handler in handlers:
                # Get the return value of the
                # default handler, and call its
                # fallback if necessary.
                if handler == 'defHandler':
                    retval = defHandler(*args, **kwargs)

                    if retval is False and fallback is not None:
                        retval = fallback(*args, **kwargs)

                else:
                    handler(*args, **kwargs)

            if self.__postEventHandler is not None:
                self.__postEventHandler(mode, evType)

            # Store the last event
            # that was processed
            self.__lastHandler = (mode, evType)

            return retval

        if len(handlers) > 0:
            log.debug('{} Handler(s) found for mode {}, event {}'.format(
                len(handlers), mode, evType))
            return handlerWrapper

        return None


    def __onMouseWheel(self, ev):
        """Called when the mouse wheel is moved.

        Delegates to a mode specific handler if one is present.
        """

        handler = self.__getHandler(ev, 'MouseWheel')
        if handler is None:
            return

        mouseLoc, canvasLoc = self.__getMouseLocation(ev)
        canvas              = ev.GetEventObject()
        wheel               = ev.GetWheelRotation()

        # wx/osx has this really useful feature
        # whereby if shift is being held down
        # (typically used for horizontal scrolling),
        # a mouse wheel direction which would have
        # produced positive values will now produce
        # negative values.
        if ev.ShiftDown() and \
           fwidgets.wxPlatform() in (fwidgets.WX_MAC_COCOA,
                                     fwidgets.WX_MAC_CARBON):
            wheel = -wheel

        log.debug('Mouse wheel event ({}) on {}'.format(
            wheel, type(canvas).__name__))

        handler(ev, canvas, wheel, mouseLoc, canvasLoc)


    def __onMouseEnter(self, ev):
        """Called when the mouse enters a canvas target.

        Delegates to a mode specific handler if one is present.
        """

        handler = self.__getHandler(ev, 'MouseEnter')
        if handler is None:
            return

        canvas              = ev.GetEventObject()
        mouseLoc, canvasLoc = self.__getMouseLocation(ev)

        log.debug('Mouse enter event on {}'.format(
            type(canvas).__name__))

        handler(ev, canvas, mouseLoc, canvasLoc)


    def __onMouseLeave(self, ev):
        """Called when the mouse leaves a canvas target.

        Delegates to a mode specific handler if one is present.
        """

        handler = self.__getHandler(ev, 'MouseLeave')
        if handler is None:
            return

        canvas              = ev.GetEventObject()
        mouseLoc, canvasLoc = self.__getMouseLocation(ev)

        log.debug('Mouse leave event on {}'.format(
            type(canvas).__name__))

        handler(ev, canvas, mouseLoc, canvasLoc)


    def __onMouseDown(self, ev):
        """Called when any mouse button is pushed.

        Delegates to a mode specific handler if one is present.
        """

        mouseLoc, canvasLoc = self.__getMouseLocation(ev)
        canvas              = ev.GetEventObject()

        # On GTK, a GLCanvas won't be given
        # focus when it is clicked on.
        canvas.SetFocus()

        # Save information about this mouse
        # down event, as it may be needed
        # by subesquent drag/up events.
        self.__mouseDownPos  = mouseLoc
        self.__canvasDownPos = canvasLoc
        self.__mouseDownMode = self.__getMode(ev)

        if self.__lastMousePos  is None: self.__lastMousePos  = mouseLoc
        if self.__lastCanvasPos is None: self.__lastCanvasPos = canvasLoc

        evType  = '{}MouseDown'.format(self.__getMouseButton(ev))
        handler = self.__getHandler(ev, evType)
        if handler is None:
            ev.Skip()
            return

        log.debug('{} event ({}, {}) on {}'.format(
            evType, mouseLoc, canvasLoc, type(canvas).__name__))

        # If a handler returns None, we
        # assume that it means True
        if handler(ev, canvas, mouseLoc, canvasLoc) is False:
            ev.Skip()

        self.__lastMousePos  = mouseLoc
        self.__lastCanvasPos = canvasLoc


    def __onMouseUp(self, ev):
        """Called when any mouse button is released.

        Delegates to a mode specific handler if one is present.
        """

        evType  = '{}MouseUp'.format(self.__getMouseButton(ev))
        handler = self.__getHandler(ev, evType, mode=self.__mouseDownMode)

        if handler is None:
            self.__mouseDownPos  = None
            self.__canvasDownPos = None
            self.__mouseDownMode = None
            ev.Skip()
            return

        canvas              = ev.GetEventObject()
        mouseLoc, canvasLoc = self.__getMouseLocation(ev)

        log.debug('{} event ({}, {}) on {}'.format(
            evType, mouseLoc, canvasLoc, type(canvas).__name__))

        if handler(ev, canvas, mouseLoc, canvasLoc) is False:
            ev.Skip()

        self.__lastMouseUpHandler = (self.__mouseDownMode, evType)
        self.__lastMouseUpPos     = mouseLoc
        self.__lastCanvasUpPos    = canvasLoc
        self.__mouseDownPos       = None
        self.__canvasDownPos      = None
        self.__mouseDownMode      = None


    def __onMouseMove(self, ev):
        """Called on mouse motion. If a mouse button is down, delegates to
        :meth:`__onMouseDrag`.

        Otherwise, delegates to a mode specific handler if one is present.
        """

        if ev.Dragging():
            self.__onMouseDrag(ev)
            return

        handler = self.__getHandler(ev, 'MouseMove')

        if handler is None:
            ev.Skip()
            return

        canvas              = ev.GetEventObject()
        mouseLoc, canvasLoc = self.__getMouseLocation(ev)

        log.debug('Mouse move event ({}, {}) on {}'.format(
            mouseLoc, canvasLoc, type(canvas).__name__))

        if handler(ev, canvas, mouseLoc, canvasLoc) is False:
            ev.Skip()

        self.__lastMousePos  = mouseLoc
        self.__lastCanvasPos = canvasLoc


    def __onMouseDrag(self, ev):
        """Called on mouse drags.

        Delegates to a mode specific handler if one is present.
        """
        ev.Skip()

        canvas              = ev.GetEventObject()
        mouseLoc, canvasLoc = self.__getMouseLocation(ev)

        evType  = '{}MouseDrag'.format(self.__getMouseButton(ev))
        handler = self.__getHandler(ev, evType, mode=self.__mouseDownMode)
        if handler is None:
            ev.Skip()
            return

        log.debug('{} event ({}, {}) on {}'.format(
            evType, mouseLoc, canvasLoc, type(canvas).__name__))

        if handler(ev, canvas, mouseLoc, canvasLoc) is False:
            ev.Skip()

        self.__lastMousePos  = mouseLoc
        self.__lastCanvasPos = canvasLoc


    def __onChar(self, ev):
        """Called on keyboard key presses.

        Delegates to a mode specific handler if one is present.
        """

        handler = self.__getHandler(ev, 'Char')
        if handler is None:
            ev.Skip()
            return

        canvas = ev.GetEventObject()
        key    = ev.GetKeyCode()

        log.debug('Keyboard event ({}) on {}'.format(
            key, type(canvas).__name__))

        if handler(ev, canvas, key) is False:
            ev.Skip()


    def __onPick(self, ev):
        """Called by the :meth:`handlePickEvent`. Delegates the event to a
        suitable handler, if one exists.
        """

        evType  = '{}MouseArtistPick'.format(self.__getMouseButton(ev))
        handler = self.__getHandler(ev, evType)
        if handler is None:
            ev.Skip()
            return

        canvas              = ev.GetEventObject()
        artist              = self.__evtManager.getPickedArtist()
        mouseLoc, canvasLoc = self.__getMouseLocation(ev)

        log.debug('Pick event ({}, {}) on {}'.format(
            mouseLoc, canvasLoc, type(canvas).__name__))

        if handler(ev, canvas, artist, mouseLoc, canvasLoc) is False:
            ev.Skip()


class CanvasPanelEventManager:
    """This class manages ``wx`` mouse/keyboard events originating from
    :class:`.SliceCanvas` instances contained within :class:`.CanvasPanel`
    views.
    """

    def __init__(self, profile):
        """Create a ``CanvasPanelEventManager``.

        :arg profile: The :class:`Profile` instance that owns this event
                      manager.
        """
        self.__profile = profile


    def register(self):
        """This method must be called to register event listeners on mouse/
        keyboard events. This method is called by meth :`Profile.register`.
        """
        for t in self.__profile.getEventTargets():
            t.Bind(wx.EVT_LEFT_DOWN,    self.__onEvent)
            t.Bind(wx.EVT_MIDDLE_DOWN,  self.__onEvent)
            t.Bind(wx.EVT_RIGHT_DOWN,   self.__onEvent)
            t.Bind(wx.EVT_LEFT_UP,      self.__onEvent)
            t.Bind(wx.EVT_MIDDLE_UP,    self.__onEvent)
            t.Bind(wx.EVT_RIGHT_UP,     self.__onEvent)
            t.Bind(wx.EVT_MOTION,       self.__onEvent)
            t.Bind(wx.EVT_MOUSEWHEEL,   self.__onEvent)
            t.Bind(wx.EVT_ENTER_WINDOW, self.__onEvent)
            t.Bind(wx.EVT_LEAVE_WINDOW, self.__onEvent)
            t.Bind(wx.EVT_CHAR,         self.__onEvent)


    def deregister(self):
        """This method de-registers mouse/keybouard event listeners. This
        method is called by the :meth:`Profile.deregister` method.
        """
        for t in self.__profile.getEventTargets():
            t.Unbind(wx.EVT_LEFT_DOWN)
            t.Unbind(wx.EVT_MIDDLE_DOWN)
            t.Unbind(wx.EVT_RIGHT_DOWN)
            t.Unbind(wx.EVT_LEFT_UP)
            t.Unbind(wx.EVT_MIDDLE_UP)
            t.Unbind(wx.EVT_RIGHT_UP)
            t.Unbind(wx.EVT_MOTION)
            t.Unbind(wx.EVT_MOUSEWHEEL)
            t.Unbind(wx.EVT_ENTER_WINDOW)
            t.Unbind(wx.EVT_LEAVE_WINDOW)
            t.Unbind(wx.EVT_CHAR)


    def getMouseLocation(self, ev):
        """Returns two tuples; the first contains the x/y coordinates of the
        given :class:`wx.MouseEvent`, and the second contains the
        corresponding x/y/z display space coordinates.
        """

        mx, my    = ev.GetPosition()
        canvas    = ev.GetEventObject()
        w, h      = canvas.GetClientSize()
        my        = h - my

        canvasPos = canvas.canvasToWorld(mx, my)

        return (mx, my), canvasPos


    def __onEvent(self, ev):
        """Event handler. Passes the event to :class:`Profile.handleEvent`. """
        self.__profile.handleEvent(ev)


class PlotPanelEventManager:
    """This class manages events originating from ``matplotlib`` ``Canvas``
    objects contained within :class:`.PlotPanel` views.


    .. note:: This class has no support for ``figure_enter_event``,
              ``figure_leave_event``, ``axis_enter_event``, or
              ``axis_leave_event`` events. There appears to be some bugs
              lurking in the ``matplotlib/backend_bases.py``  file (something
              to do with the ``LocationEvent.lastevent`` property) which
              cause FSLeyes to seg-fault on ``figure_enter_event`` events.
    """


    def __init__(self, profile):
        """Create a ``PlotPanelEventManager``.

        :arg profile: The :class:`Profile` instance that owns this event
                      manager.
        """
        self.__profile       = profile
        self.__lastEvent     = None
        self.__lastLocEvent  = None
        self.__lastPickEvent = None
        self.__cids          = {}
        self.__eventTypes    = [
            'button_press_event',
            'button_release_event',
            'motion_notify_event',
            'pick_event',
            'scroll_event',
            'key_press_event',
            'key_release_event']


    def register(self):
        """Register listeners on all relevant ``matplotlib`` events. This
        method is called by :meth:`Profile.register`.
        """

        for target in self.__profile.getEventTargets():
            for ev in self.__eventTypes:
                self.__cids[ev] = target.mpl_connect(ev, self.__onEvent)


    def deregister(self):
        """De-register listeners on all relevant ``matplotlib`` events. This
        method is called by :meth:`Profile.deregister`.
        """

        for target in self.__profile.getEventTargets():
            for ev in self.__eventTypes:
                target.mpl_disconnect(self.__cids[ev])
        self.__cids = {}


    def getMouseLocation(self, ev=None):
        """Returns two tuples; the first contains the x/y coordinates of the
        most recent ``matplotlib`` event, and the second contains the
        corresponding x/y data coordinates.

        If an event has not yet occurred, or if the mouse position is not on
        any event target, this method returns ``(None, None)``.

        :arg ev: Ignored.
        """

        mplev = self.__lastLocEvent

        if mplev       is None: return None, None
        if mplev.xdata is None: return None, None
        if mplev.ydata is None: return None, None

        mousex, mousey = mplev.x,     mplev.y
        datax,  datay  = mplev.xdata, mplev.ydata

        return (mousex, mousey), (datax, datay)


    def getPickedArtist(self):
        """Returns the ``matplotlib.Artist`` that was most recently picked
        (clicked on) by the user.
        """

        mplev = self.__lastPickEvent

        if mplev is None: return None
        else:             return mplev.artist


    def getMplEvent(self):
        """Returns the most recent ``matplotlib`` event that occurred, or
        ``None`` if no events have occurred yet.
        """
        return self.__lastEvent


    def __onEvent(self, ev):
        """Handler for ``matplotlib`` events. Passes the corresponding ``wx``
        event to the :meth:`Profile.handleEvent` method.
        """

        if ev.name not in self.__eventTypes:
            return

        self.__lastEvent = ev

        islocev  = isinstance(ev, mplbackend.LocationEvent)
        ispickev = isinstance(ev, mplbackend.PickEvent)

        if islocev:  self.__lastLocEvent  = ev
        if ispickev: self.__lastPickEvent = ev

        if ispickev: self.__profile.handlePickEvent(ev.guiEvent)
        else:        self.__profile.handleEvent(    ev.guiEvent)



#!/usr/bin/env python
#
# togglecontrolpanel.py - The ToggleControlPanelAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ToggleControlPanelAction` class,
an action intended for use by :class:`.ViewPanel` classes to manage
the addition/removal of control panels.
"""


import wx.lib.agw.aui as aui

from . import base


class ToggleControlPanelAction(base.ToggleAction):
    """The ``ToggleControlPanelAction`` class is an action which is
    intended to augment :class:`.ViewPanel` actions which add/remove
    a control panel.

    The ``ToggleControlPanelAction`` keeps track of the :class:`.ViewPanel`
    state, and updates its own :attr:`.ToggleAction.toggled` property whenever
    its designated control panel is added/removed from the ``ViewPanel``.

    This sounds a bit silly, but is necessary to ensure that any bound
    widgets (most likely menu items) are updated whenever the control panel
    managed by a ``ToggleControlPanelAction`` is added/removed.
    """

    def __init__(self,
                 overlayList,
                 displayCtx,
                 viewPanel,
                 cpType,
                 func=None,
                 name=None,
                 **kwargs):
        """Create a ``ToggleControlPanelAction``.

        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext`
        :arg viewPanel:   The :class:`.ViewPanel` instance.
        :arg cpType:      The type of the control panel being managed by this
                          ``ToggleControlPanelAction``.
        :arg func:        The function which toggles the control panel. If
                          not provided, a default function is used.
        :arg name:        Name of this action - defaults to ``func.__name__``.

        All other arguments will be passed to the
        :meth:`.ViewPanel.togglePanel` method.
        """

        if func is None:
            func = self.__togglePanel

        base.ToggleAction.__init__(
            self, overlayList, displayCtx, func, name=name)

        self.__viewPanel = viewPanel
        self.__cpType    = cpType
        self.__kwargs    = kwargs

        # Listen for changes to the view panel layout
        # so we can detect when the user closes our
        # control panel. The granularity of AuiManager
        # event notifications is somewhat coarse, so
        # this callback will be called whenever the
        # perspective changes. This includes sash
        # resizes :(
        viewPanel.events.register(self.name,
                                  self.__viewPanelChanged,
                                  topic='aui_perspective')


    def __togglePanel(self, *args, **kwargs):
        """Default action to run if a ``func`` wasn't specified. Calls
        :class:`.ViewPanel.togglePanel`,
        """
        self.viewPanel.togglePanel(self.__cpType,
                                   *args,
                                   **kwargs,
                                   **self.__kwargs)


    @property
    def viewPanel(self):
        """Returns a reference to the :class:`.ViewPanel` that is associated
        with this action.
        """
        return self.__viewPanel


    def destroy(self):
        """Must be called when this ``ToggleControlPanelAction`` is no longer
        used. Clears references, and calls the base-class ``destroy`` method.
        """

        if self.destroyed:
            return

        if not self.__viewPanel.destroyed:
            self.__viewPanel.events.deregister(
                self.name, topic='aui_perspective')
        self.__viewPanel = None

        base.ToggleAction.destroy(self)


    def __viewPanelChanged(self, *a):
        """Called whenever a control panel is added to/removed from the
        :class:`.ViewPanel` that owns this ``ToggleControlPanelAction``.
        Updates the :attr:`.ToggleAction.toggled` attribute of this action.
        """

        if self.destroyed:
            return

        controlPanels = self.__viewPanel.getPanels()
        controlPanels = [type(cp) for cp in controlPanels]
        self.toggled = self.__cpType in controlPanels



#!/usr/bin/env python
#
# locationpanel.py - provides the LocationPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LocationPanel` class, a *FSLeyes control*
panel which shows information about the current display location.
"""


import logging
import itertools as it

import wx
import wx.html as wxhtml

import numpy as np

import fsl.transform.affine           as affine
import fsl.utils.settings             as fslsettings
import fsl.data.image                 as fslimage
import fsl.data.mesh                  as fslmesh
import fsl.data.constants             as constants

import fsleyes_props                  as props
import fsleyes_widgets.floatspin      as floatspin
import fsleyes_widgets.notebook       as notebook
import fsleyes_widgets.elistbox       as elistbox
import fsleyes_widgets.utils.status   as status
import fsleyes_widgets.utils.typedict as td

import fsleyes.controls.controlpanel  as ctrlpanel
import fsleyes.views.canvaspanel      as canvaspanel
import fsleyes.panel                  as fslpanel
import fsleyes.strings                as strings


log = logging.getLogger(__name__)


class LocationPanel(ctrlpanel.ControlPanel):
    """The ``LocationPanel`` is a panel which contains controls allowing the
    user to view and modify the :attr:`.DisplayContext.location` property.

    A ``LocationPanel`` is intended to be contained within a
    :class:`.CanvasPanel`, and looks something like this:

    .. image:: images/locationpanel.png
       :scale: 50%
       :align: center

    By default, the ``LocationPanel`` contains a notebook with two pages:

      - The :class:`LocationInfoPanel` contains information about the currently
        displayed location, and controls allowing the user to change the
        location.

      - The :class:`LocationHistoryPanel` contains a list of previously visited
        locations, and allows the user to revisit those locations.

    The history panel is optional - if the ``showHistory`` parameter to
    ``__init__`` is ``False`` then only the information panel will be shown.
    """


    @staticmethod
    def supportedViews():
        """The ``LocationPanel`` is restricted for use with
        :class:`.CanvasPanel` views.
        """
        return [canvaspanel.CanvasPanel]


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'location' : wx.BOTTOM}


    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 viewPanel,
                 showHistory=False):
        """Creat a ``LocationPanel``.

        :arg parent:      The :mod:`wx` parent object, assumed to be a
                          :class:`.CanvasPanel`.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg viewPanel:   The :class:`.ViewPanel` instance.

        :arg showHistory: Defaults to ``False``. If ``True``, create and
                          display a :class:`LocationHistoryPanel`.
        """
        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, viewPanel)

        # only use a notebook if showing history panel
        if showHistory:
            self.__notebook = notebook.Notebook(self,
                                                style=wx.LEFT | wx.VERTICAL,
                                                border=0)
            subparent       = self.__notebook
        else:
            subparent       = self
            self.__notebook = None

        # We always create an info panel
        self.__info = LocationInfoPanel(
            subparent, overlayList, displayCtx, viewPanel.frame)

        # We don't always create a history panel
        if showHistory:
            self.__history = LocationHistoryPanel(
                subparent, overlayList, displayCtx, viewPanel.frame, parent)
            self.__notebook.AddPage(self.__info,
                                    strings.labels[self, 'info'])
            self.__notebook.AddPage(self.__history,
                                    strings.labels[self, 'history'])
        else:
            self.__history = None

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)

        if showHistory:
            self.__sizer.Add(self.__notebook, flag=wx.EXPAND, proportion=1)
        else:
            self.__sizer.Add(self.__info, flag=wx.EXPAND, proportion=1)

        self.SetSizer(self.__sizer)
        self.Layout()


    def destroy(self):
        """Must be called when this ``LocationPanel`` is no longer needed. """

        if self.__history is not None:
            self.__history.destroy()

        self.__info.destroy()
        self.__info    = None
        self.__history = None

        ctrlpanel.ControlPanel.destroy(self)


class LocationInfoPanel(fslpanel.FSLeyesPanel):
    """The ``LocationInfoPanel`` is a panel which is embedded in the
    :class:`LocationPanel`, and which contains controls allowing the user to
    view and modify the :attr:`.DisplayContext.location` property.


    The ``LocationInfoPanel`` contains two main sections:

      - A collection of controls which show the current
        :attr:`.DisplayContext.location`

      - A ``wx.html.HtmlWindow`` which displays information about the current
        :attr:`.DisplayContext.location` for all overlays in the
        :class:`.OverlayList`.


    **NIFTI overlays**


    The ``LocationInfoPanel`` is primarily designed to work with
    :class:`.Image` overlays. If the :attr:`.DisplayContext.selectedOverlay`
    is an :class:`.Image`, or has an associated reference image (see
    :meth:`.DisplayOpts.referenceImage`), the ``LocationInfoPanel`` will
    display the current :class:`.DisplayContext.location` in both the the
    voxel coordinates and world coordinates of the ``Image`` instance.


    **Other overlays**


    If the :attr:`.DisplayContext.selectedOverlay` is not an :class:`.Image`,
    or does not have an associated reference image, the ``LocationInfoPanel``
    will display the current :attr:`.DisplayContext.location` as-is (i.e. in
    the display coordinate system); furthermore, the voxel location controls
    will be disabled.


    **Location updates**


    The :data:`DISPLAYOPTS_BOUNDS` and :data:`DISPLAYOPTS_INFO` dictionaries
    contain lists of property names that the :class:`.LocationInfoPanel`
    listens on for changes, so it knows when the location widgets, and
    information about the currenty location, need to be refreshed. For
    example, when the :attr`.NiftiOpts.volume` property of a :class:`.Nifti`
    overlay changes, the volume index, and potentially the overlay
    information, needs to be updated.
    """


    voxelLocation = props.Point(ndims=3, real=False)
    """If the currently selected overlay is a :class:`.Image` instance , this
    property tracks the current :attr:`.DisplayContext.location` in voxel
    coordinates.
    """


    worldLocation = props.Point(ndims=3, real=True)
    """For :class:`.Image` overlays, this property tracks the current
    :attr:`.DisplayContext.location` in the image world coordinates. For other
    overlay types, this property tracks the current location in display
    coordinates.
    """


    def __init__(self, parent, overlayList, displayCtx, frame):

        fslpanel.FSLeyesPanel.__init__(self,
                                       parent,
                                       overlayList,
                                       displayCtx,
                                       frame,
                                       kbFocus=True)

        # Whenever the selected overlay changes,
        # a reference to it and its DisplayOpts
        # instance is stored, as property listeners
        # are registered on it (and need to be
        # de-registered later on).
        self.__registeredOverlay = None
        self.__registeredDisplay = None
        self.__registeredOpts    = None

        self.__column1 = wx.Panel(self)
        self.__column2 = wx.Panel(self)
        self.__info    = wxhtml.HtmlWindow(self)

        # HTMLWindow does not use
        # the parent font by default,
        # so we force it to at least
        # have the parent font size
        self.__info.SetStandardFonts(self.GetFont().GetPointSize())

        self.__worldLabel = wx.StaticText(
            self.__column1, label=strings.labels[self, 'worldLocation'])
        self.__volumeLabel = wx.StaticText(
            self.__column1, label=strings.labels[self, 'volume'])
        self.__voxelLabel = wx.StaticText(
            self.__column2, label=strings.labels[self, 'voxelLocation'])

        worldX, worldY, worldZ = props.makeListWidgets(
            self.__column1,
            self,
            'worldLocation',
            slider=False,
            spin=True,
            showLimits=False,
            mousewheel=True)

        voxelX, voxelY, voxelZ = props.makeListWidgets(
            self.__column2,
            self,
            'voxelLocation',
            slider=False,
            spin=True,
            spinWidth=7,
            showLimits=False,
            mousewheel=True)

        self.__worldX = worldX
        self.__worldY = worldY
        self.__worldZ = worldZ
        self.__voxelX = voxelX
        self.__voxelY = voxelY
        self.__voxelZ = voxelZ
        self.__volume = floatspin.FloatSpinCtrl(
            self.__column2,
            width=7,
            style=floatspin.FSC_MOUSEWHEEL | floatspin.FSC_INTEGER)

        self.__column1Sizer = wx.BoxSizer(wx.VERTICAL)
        self.__column2Sizer = wx.BoxSizer(wx.VERTICAL)
        self.__sizer        = wx.BoxSizer(wx.HORIZONTAL)

        self.__column1Sizer.Add(self.__worldLabel,  flag=wx.EXPAND)
        self.__column1Sizer.Add(self.__worldX,      flag=wx.EXPAND)
        self.__column1Sizer.Add(self.__worldY,      flag=wx.EXPAND)
        self.__column1Sizer.Add(self.__worldZ,      flag=wx.EXPAND)
        self.__column1Sizer.Add(self.__volumeLabel, flag=wx.ALIGN_RIGHT)

        self.__column2Sizer.Add(self.__voxelLabel, flag=wx.EXPAND)
        self.__column2Sizer.Add(self.__voxelX,     flag=wx.EXPAND)
        self.__column2Sizer.Add(self.__voxelY,     flag=wx.EXPAND)
        self.__column2Sizer.Add(self.__voxelZ,     flag=wx.EXPAND)
        self.__column2Sizer.Add(self.__volume,     flag=wx.EXPAND)

        self.__sizer.Add(self.__column1, flag=wx.EXPAND)
        self.__sizer.Add((5, -1))
        self.__sizer.Add(self.__column2, flag=wx.EXPAND)
        self.__sizer.Add((5, -1))
        self.__sizer.Add(self.__info,    flag=wx.EXPAND, proportion=1)

        self.__column1.SetSizer(self.__column1Sizer)
        self.__column2.SetSizer(self.__column2Sizer)
        self          .SetSizer(self.__sizer)

        self.overlayList.addListener('overlays',
                                     self.name,
                                     self.__selectedOverlayChanged)
        self.displayCtx .addListener('selectedOverlay',
                                     self.name,
                                     self.__selectedOverlayChanged)
        self.displayCtx .addListener('overlayOrder',
                                     self.name,
                                     self.__overlayOrderChanged)
        self.displayCtx .addListener('location',
                                     self.name,
                                     self.__displayLocationChanged)
        self.addListener(            'voxelLocation',
                                     self.name,
                                     self.__voxelLocationChanged)
        self.addListener(            'worldLocation',
                                     self.name,
                                     self.__worldLocationChanged)

        self.__selectedOverlayChanged()

        self.__worldLabel .SetMinSize(self.__calcWorldLabelMinSize())
        self.__voxelLabel .SetMinSize(self.__voxelLabel .GetBestSize())
        self.__volumeLabel.SetMinSize(self.__volumeLabel.GetBestSize())
        self.__column1    .SetMinSize(self.__column1    .GetBestSize())
        self.__column2    .SetMinSize(self.__column2    .GetBestSize())
        self.__info       .SetMinSize((100, 100))

        # Keyboard navigation - see FSLeyesPanel
        self.setNavOrder((self.__worldX,
                          self.__worldY,
                          self.__worldZ,
                          self.__voxelX,
                          self.__voxelY,
                          self.__voxelZ,
                          self.__volume))

        self.Layout()

        self.__minSize = self.__sizer.GetMinSize()
        self.SetMinSize(self.__minSize)


    def destroy(self):
        """Must be called when this ``LocationInfoPanel`` is no longer needed.
        Removes property listeners and calls :meth:`.FSLeyesPanel.destroy`.
        """

        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.displayCtx .removeListener('location',        self.name)

        self.__deregisterOverlay()

        fslpanel.FSLeyesPanel.destroy(self)


    def GetMinSize(self):
        """Returns the minimum size for this ``LocationInfoPanel``.

        Under Linux/GTK, the ``wx.agw.lib.aui`` layout manager seems to
        arbitrarily adjust the minimum sizes of some panels. Therefore, The
        minimum size of the ``LocationInfoPanel`` is calculated in
        :meth:`__init__`, and is fixed.
        """
        return self.__minSize


    def DoGetBestClientSize(self):
        """Returns the best size for this ``LocationInfoPanel``.
        """
        return self.__minSize


    def __calcWorldLabelMinSize(self):
        """Calculates the minimum size that the world label (the label which
        shows the coordinate space of the currently selected overlay) needs.
        Called by the :meth:`__init__` method.

        The world label displays different things depending on the currently
        selected overlay. But we want it to be a fixed size. So this method
        calculates the size of all possible values that the world label will
        display, and returns the maximum size. This is then used as the
        minimum size for the world label.
        """

        dc = wx.ClientDC(self.__worldLabel)

        width, height = 0, 0

        labelPref = strings.labels[self, 'worldLocation']
        labelSufs = [
            strings.anatomy['Nifti',
                            'space',
                            constants.NIFTI_XFORM_UNKNOWN],
            strings.anatomy['Nifti',
                            'space',
                            constants.NIFTI_XFORM_SCANNER_ANAT],
            strings.anatomy['Nifti',
                            'space',
                            constants.NIFTI_XFORM_ALIGNED_ANAT],
            strings.anatomy['Nifti',
                            'space',
                            constants.NIFTI_XFORM_TALAIRACH],
            strings.anatomy['Nifti',
                            'space',
                            constants.NIFTI_XFORM_MNI_152],
            strings.labels[self, 'worldLocation', 'unknown']
        ]

        for labelSuf in labelSufs:

            w, h = dc.GetTextExtent(labelPref + labelSuf)

            if w > width:  width  = w
            if h > height: height = h

        return width + 5, height


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` or
        :class:`.OverlayList` is changed. Registered with the new overlay,
        and refreshes the ``LocationInfoPanel`` interface accordingly.
        """

        self.__deregisterOverlay()

        if len(self.overlayList) == 0:
            self.__updateWidgets()
            self.__updateLocationInfo()

        else:
            self.__registerOverlay()
            self.__updateWidgets()
            self.__displayLocationChanged()


    def __registerOverlay(self):
        """Registers property listeners with the :class:`.Display` and
        :class:`.DisplayOpts` instances associated with the currently
        selected overlay.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        display = self.displayCtx.getDisplay(overlay)
        opts    = display.opts

        self.__registeredOverlay = overlay
        self.__registeredDisplay = display
        self.__registeredOpts    = opts

        # The properties that we need to
        # listen for are specified in the
        # DISPLAYOPTS_BOUNDS and
        # DISPLAYOPTS_INFO dictionaries.
        boundPropNames = DISPLAYOPTS_BOUNDS.get(opts, [], allhits=True)
        infoPropNames  = DISPLAYOPTS_INFO  .get(opts, [], allhits=True)
        boundPropNames = it.chain(*boundPropNames)
        infoPropNames  = it.chain(*infoPropNames)

        # DisplayOpts instances get re-created
        # when an overlay type is changed, so
        # we need to re-register when this happens.
        display.addListener('overlayType',
                            self.name,
                            self.__selectedOverlayChanged)

        for n in boundPropNames:
            opts.addListener(n, self.name, self.__boundsOptsChanged)
        for n in infoPropNames:
            opts.addListener(n, self.name, self.__infoOptsChanged)

        # Enable the volume widget if the
        # overlay is a NIFTI image with more
        # than three dimensions, and bind
        # the widget to the volume property
        # of  the associated NiftiOpts instance
        if isinstance(overlay, fslimage.Nifti) and overlay.ndim > 3:

            props.bindWidget(self.__volume,
                             opts,
                             'volume',
                             floatspin.EVT_FLOATSPIN)

            opts.addListener('volumeDim', self.name, self.__volumeDimChanged)

            self.__volume     .Enable()
            self.__volumeLabel.Enable()
            self.__volumeDimChanged()

        # Or, if the overlay is a mesh which
        # has some time series data associated
        # with it
        elif isinstance(overlay, fslmesh.Mesh):

            props.bindWidget(self.__volume,
                             opts,
                             'vertexDataIndex',
                             floatspin.EVT_FLOATSPIN)

            opts.addListener('vertexData', self.name, self.__vertexDataChanged)
            self.__vertexDataChanged()

        else:
            self.__volume.SetRange(0, 0)
            self.__volume.SetValue(0)
            self.__volume.Disable()


    def __deregisterOverlay(self):
        """De-registers property listeners with the :class:`.Display` and
        :class:`.DisplayOpts` instances associated with the previously
        registered overlay.
        """

        opts    = self.__registeredOpts
        display = self.__registeredDisplay
        overlay = self.__registeredOverlay

        if overlay is None:
            return

        self.__registeredOpts    = None
        self.__registeredDisplay = None
        self.__registeredOverlay = None

        boundPropNames = DISPLAYOPTS_BOUNDS.get(opts, [], allhits=True)
        infoPropNames  = DISPLAYOPTS_INFO  .get(opts, [], allhits=True)
        boundPropNames = it.chain(*boundPropNames)
        infoPropNames  = it.chain(*infoPropNames)

        if display is not None:
            display.removeListener('overlayType', self.name)

        for p in boundPropNames: opts.removeListener(p, self.name)
        for p in infoPropNames:  opts.removeListener(p, self.name)

        if isinstance(overlay, fslimage.Nifti) and overlay.ndim > 3:
            props.unbindWidget(self.__volume,
                               opts,
                               'volume',
                               floatspin.EVT_FLOATSPIN)
            opts.removeListener('volumeDim', self.name)

        elif isinstance(overlay, fslmesh.Mesh):
            props.unbindWidget(self.__volume,
                               opts,
                               'vertexDataIndex',
                               floatspin.EVT_FLOATSPIN)
            opts.removeListener('vertexData', self.name)


    def __volumeDimChanged(self, *a):
        """Called when the selected overlay is a :class:`.Nifti`, and its
        :attr:`.NiftiOpts.volumeDim` property changes. Updates the volume
        widget.
        """
        overlay = self.__registeredOverlay
        opts    = self.__registeredOpts
        volume  = opts.volume
        vdim    = opts.volumeDim + 3

        self.__volume.SetRange(0, overlay.shape[vdim] - 1)
        self.__volume.SetValue(volume)
        self.__infoOptsChanged()


    def __vertexDataChanged(self, *a):
        """Called when the selected overlay is a :class:`.Mesh`, and
        its :attr:`.MeshOpts.vertexData` property changes. Updates the volume
        widget.
        """

        opts    = self.__registeredOpts
        vd      = opts.getVertexData()
        vdi     = opts.vertexDataIndex
        enabled = vd is not None and vd.shape[1] > 1

        self.__volume     .Enable(enabled)
        self.__volumeLabel.Enable(enabled)

        if enabled:
            self.__volume.SetRange(0, vd.shape[1] - 1)
            self.__volume.SetValue(vdi)

        self.__infoOptsChanged()


    def __boundsOptsChanged(self, *a):
        """Called when a :class:`.DisplayOpts` property associated
        with the currently selected overlay, and listed in the
        :data:`DISPLAYOPTS_BOUNDS` dictionary, changes. Refreshes the
        ``LocationInfoPanel`` interface accordingly.
        """
        self.__updateWidgets()
        self.__displayLocationChanged()


    def __infoOptsChanged(self, *a):
        """Called when a :class:`.DisplayOpts` property associated
        with the currently selected overlay, and listed in the
        :data:`DISPLAYOPTS_INFO` dictionary, changes. Refreshes the
        ``LocationInfoPanel`` interface accordingly.
        """
        self.__displayLocationChanged()


    def __overlayOrderChanged(self, *a):
        """Called when the :attr:`.DisplayContext.overlayOrder` changes,
        Refreshes the information panel.
        """
        self.__displayLocationChanged()


    def __updateWidgets(self):
        """Called by the :meth:`__selectedOverlayChanged` and
        :meth:`__displayOptsChanged` methods.  Enables/disables the
        voxel/world location and volume controls depending on the currently
        selected overlay (or reference image).
        """

        overlay = self.__registeredOverlay
        opts    = self.__registeredOpts

        if overlay is not None: refImage = opts.referenceImage
        else:                   refImage = None

        haveRef = refImage is not None

        self.__voxelX     .Enable(haveRef)
        self.__voxelY     .Enable(haveRef)
        self.__voxelZ     .Enable(haveRef)
        self.__voxelLabel .Enable(haveRef)

        ######################
        # World location label
        ######################

        label = strings.labels[self, 'worldLocation']

        if haveRef: label += strings.anatomy[refImage,
                                             'space',
                                             refImage.getXFormCode()]
        else:       label += strings.labels[ self,
                                             'worldLocation',
                                             'unknown']

        self.__worldLabel.SetLabel(label)

        ####################################
        # Voxel/world location widget limits
        ####################################

        # Figure out the limits for the
        # voxel/world location widgets
        if haveRef:
            opts     = self.displayCtx.getOpts(refImage)
            v2w      = opts.getTransform('voxel', 'world')
            shape    = refImage.shape[:3]
            vlo      = [0, 0, 0]
            vhi      = np.array(shape) - 1
            wlo, whi = affine.axisBounds(shape, v2w)
            wstep    = refImage.pixdim[:3]
        else:
            vlo     = [0, 0, 0]
            vhi     = [0, 0, 0]
            wbounds = self.displayCtx.bounds[:]
            wlo     = wbounds[0::2]
            whi     = wbounds[1::2]
            wstep   = [1, 1, 1]

        log.debug('Setting voxelLocation limits: {} - {}'.format(vlo, vhi))
        log.debug('Setting worldLocation limits: {} - {}'.format(wlo, whi))

        # Update the voxel and world location limits,
        # but don't trigger a listener callback, as
        # this would change the display location.
        widgets = [self.__worldX, self.__worldY, self.__worldZ]
        with props.suppress(self, 'worldLocation'), \
             props.suppress(self, 'voxelLocation'):

            for i in range(3):
                self.voxelLocation.setLimits(i, vlo[i], vhi[i])
                self.worldLocation.setLimits(i, wlo[i], whi[i])
                widgets[i].SetIncrement(wstep[i])


    def __displayLocationChanged(self, *a):
        """Called when the :attr:`.DisplayContext.location` changes.
        Propagates the change on to the :attr:`voxelLocation`
        and :attr:`worldLocation` properties.

        .. note:: Because the :attr:`.DisplayContext.location`,
                  :attr:`voxelLocation` and :attr:`worldLocation` properties
                  are all linked through property listeners (see
                  :meth:`props.HasProperties.addListener`), we need to be a
                  bit careful to avoid circular updates. Therefore, each of
                  the :meth:`__displayLocationChanged`,
                  :meth:`__worldLocationChanged` and
                  :meth:`__voxelLocationChanged` methods use the
                  :meth:`__prePropagate`, :meth:`__propagate`, and
                  :meth:`__postPropagate` methods to propagate changes
                  between the three location properties.
        """

        if not self or self.destroyed:
            return

        if len(self.overlayList) == 0:       return
        if self.__registeredOverlay is None: return

        self.__prePropagate()
        self.__propagate('display', 'voxel')
        self.__propagate('display', 'world')
        self.__postPropagate()
        self.__updateLocationInfo()


    def __worldLocationChanged(self, *a):
        """Called when the :attr:`worldLocation` changes.  Propagates the
        change on to the :attr:`voxelLocation` and
        :attr:`.DisplayContext.location` properties.
        """

        if len(self.overlayList) == 0:       return
        if self.__registeredOverlay is None: return

        self.__prePropagate()
        self.__propagate('world', 'voxel')
        self.__propagate('world', 'display')
        self.__postPropagate()
        self.__updateLocationInfo()


    def __voxelLocationChanged(self, *a):
        """Called when the :attr:`voxelLocation` changes.  Propagates the
        change on to the :attr:`worldLocation` and
        :attr:`.DisplayContext.location` properties.
        """

        if len(self.overlayList) == 0:       return
        if self.__registeredOverlay is None: return

        self.__prePropagate()
        self.__propagate('voxel', 'world')
        self.__propagate('voxel', 'display')
        self.__postPropagate()
        self.__updateLocationInfo()


    def __prePropagate(self):
        """Called by the :meth:`__displayLocationChanged`,
        :meth:`__worldLocationChanged` and :meth:`__voxelLocationChanged`
        methods.

        Disables notification of all location property listeners, so
        circular updates do not occur.
        """

        self           .disableNotification('voxelLocation')
        self           .disableNotification('worldLocation')
        self.displayCtx.disableListener(    'location', self.name)

        self.Freeze()


    def __propagate(self, source, target):
        """Called by the :meth:`__displayLocationChanged`,
        :meth:`__worldLocationChanged` and :meth:`__voxelLocationChanged`
        methods. Copies the coordinates from the ``source`` location to the
        ``target`` location. Valid values for the ``source`` and ``target``
        are:

        =========== ==============================================
        ``display`` The :attr:`.DisplayContext.location` property.
        ``voxel``   The :attr:`voxelLocation` property.
        ``world``   The :attr:`worldLocation` property.
        =========== ==============================================
        """

        if   source == 'display': coords = self.displayCtx.location.xyz
        elif source == 'voxel':   coords = self.voxelLocation.xyz
        elif source == 'world':   coords = self.worldLocation.xyz

        refImage = self.__registeredOpts.referenceImage

        if refImage is not None:
            opts    = self.displayCtx.getOpts(refImage)
            xformed = opts.transformCoords([coords],
                                           source,
                                           target,
                                           vround=target == 'voxel')[0]
        else:
            xformed = coords

        log.debug('Updating location ({} {} -> {} {})'.format(
            source, coords, target, xformed))

        if   target == 'display': self.displayCtx.location.xyz = xformed
        elif target == 'voxel':   self.voxelLocation      .xyz = xformed
        elif target == 'world':   self.worldLocation      .xyz = xformed


    def __postPropagate(self):
        """Called by the :meth:`__displayLocationChanged`,
        :meth:`__worldLocationChanged` and :meth:`__voxelLocationChanged`
        methods.

        Re-enables the property listeners that were disabled by the
        :meth:`__postPropagate` method.
        """
        self           .enableNotification('voxelLocation')
        self           .enableNotification('worldLocation')
        self.displayCtx.enableListener(    'location', self.name)

        self.Thaw()
        self.Refresh()
        self.Update()


    def __updateLocationInfo(self):
        """Called whenever the :attr:`.DisplayContext.location` changes.
        Updates the HTML panel which displays information about all overlays
        in the :class:`.OverlayList`.
        """

        if len(self.overlayList) == 0 or self.__registeredOverlay is None:
            self.__info.SetPage('')
            return

        # Reverse the overlay order so they
        # are ordered the same on the info
        # page as in the overlay list panel
        displayCtx = self.displayCtx
        overlays   = reversed(displayCtx.getOrderedOverlays())
        selOvl     = displayCtx.getSelectedOverlay()
        lines      = []

        dswarn = self.__genDisplaySpaceWarning()
        if dswarn is not None:
            fmt = '<span style="color: #ff0000"><b>{}</b></span>'
            lines.append(fmt.format(dswarn))

        for overlay in overlays:

            display = displayCtx.getDisplay(overlay)
            opts    = display.opts

            if not display.enabled:
                continue

            info = None
            title = '<b>{}</b>'.format(display.name)

            # For mesh overlays, if the current location
            # corresponds to a vertex, show some info
            # about that vertex
            if isinstance(overlay, fslmesh.Mesh):
                info = self.__genMeshInfo(overlay, opts)
            elif isinstance(overlay, fslimage.Image):
                info = self.__genImageInfo(overlay, opts)
            else:
                info = '{}'.format(strings.labels[self, 'noData'])

            # Indent info for unselected overlays,
            # to make the info for the selected
            # overlay a bit more obvious.
            colourFmt = '<span style="color: #6060ff">{}</span>'
            if overlay is selOvl:
                title = colourFmt.format(title)
                if info is not None:
                    info = colourFmt.format(info)

            lines.append(title)

            if info is not None:
                lines.append(info)

        self.__info.SetPage('<br>'.join(lines))
        self.__info.Refresh()


    def __genDisplaySpaceWarning(self):
        """Generate a warning if images with different orientations and/or
        fields-of-view are loaded.
        """
        images  = [o for o in self.overlayList
                   if isinstance(o, fslimage.Image)]

        for i in images[1:]:
            if not i.sameSpace(images[0]):
                return strings.messages[self, 'displaySpaceWarning']

        return None


    def __genMeshInfo(self, ovl, opts):
        """Generate an info line for the given :class:`.Mesh` overlay. """
        vidx  = opts.getVertex()
        vd    = opts.getVertexData()
        vdidx = opts.vertexDataIndex

        if vidx is None:
            info = '[no vertex]'

        else:
            # some vertex data has been
            # loaded for this mesh.
            if vd is not None:

                # time series/multiple data points per
                # vertex - display the time/data index
                # as well
                if vd.shape[1] > 1:
                    info = '[{}, {}]: {}'.format(vidx,
                                                 vdidx,
                                                 vd[vidx, vdidx])

                # Only one scalar value per vertex -
                # don't bother showing the vertex
                # data index
                else:
                    info = '[{}]: {}'.format(vidx, vd[vidx, vdidx])

            else:
                info = '[{}]'.format(vidx)
        return info


    def __genImageInfo(self, ovl, opts):
        """Generate an info line for the given :class:`.Image` overlay. """

        vloc = opts.getVoxel()

        if vloc is not None:
            vloc = tuple(int(v) for v in vloc)
            vloc = opts.index(vloc)
            vval = ovl[vloc]
            vloc = ' '.join(map(str, vloc))

            if not np.isscalar(vval):
                vval = np.asscalar(vval)

            if opts.overlayType == 'label':
                lbl = opts.lut.get(int(vval))
                if lbl is None: lbl = 'no label'
                else:           lbl = lbl.name
                info = '[{}]: {} / {}'.format(vloc, vval, lbl)

            elif opts.overlayType == 'complex':
                if opts.component in ('real', 'imag'):
                    info = '[{}]: {}'.format(vloc, vval)
                else:
                    cval = opts.getComponent(vval)
                    info = '[{}]: {} ({}: {})'.format(
                        vloc, vval, opts.component, cval)
            else:
                info = '[{}]: {}'.format(vloc, vval)

        else:
            info = strings.labels[self, 'outOfBounds']
        return info


DISPLAYOPTS_BOUNDS = td.TypeDict({
    'DisplayOpts' : ['bounds'],
    'MeshOpts'    : ['refImage'],
})
"""Different :class:`.DisplayOpts` types have different properties which
affect the current overlay bounds.  Therefore, when the current overlay
changes (as dictated by the :attr:`.DisplayContext.selectedOverlay`
property),the :meth:`__registerOverlay` method registers property
listeners on the properties specified in this dictionary.
"""


DISPLAYOPTS_INFO = td.TypeDict({
    'NiftiOpts'   : ['volume'],
    'MeshOpts'    : ['vertexDataIndex'],
    'ComplexOpts' : ['component']
})
"""Different :class:`.DisplayOpts` types have different properties which
affect the current overlay location information.  Therefore, when the current
overlay changes the :meth:`__registerOverlay` method registers property
listeners on the properties specified in this dictionary.
"""


class LocationHistoryPanel(fslpanel.FSLeyesPanel):
    """The ``LocationHistoryPanel`` is a panel which is embedded in the
    :class:`LocationPanel`, and which contains a list of locations
    previously visited by the user.
    """

    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 frame,
                 canvasPanel,
                 limit=500):
        """Create a ``LocationHistoryPanel``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg frame:       The :class:`.FSLeyesFrame` instance.

        :arg canvasPanel: The :class:`.CanvasPanel` which owns this
                          ``LocationHistoryPanel``.

        :arg limit:       Maximum number of locations to save before dropping
                          old locations.
        """
        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__profile  = None
        self.__limit    = limit
        self.__canvas   = canvasPanel
        self.__sizer    = wx.BoxSizer(wx.VERTICAL)
        self.__btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__load     = wx.Button(self, style=wx.BU_EXACTFIT)
        self.__save     = wx.Button(self, style=wx.BU_EXACTFIT)
        self.__clear    = wx.Button(self, style=wx.BU_EXACTFIT)
        self.__hint     = wx.StaticText(self)
        self.__list     = elistbox.EditableListBox(
            self,
            style=(elistbox.ELB_REVERSE   |
                   elistbox.ELB_NO_ADD    |
                   elistbox.ELB_NO_REMOVE |
                   elistbox.ELB_NO_MOVE   |
                   elistbox.ELB_EDITABLE  |
                   elistbox.ELB_SCROLL_BUTTONS))

        self.__load .SetLabel(strings.labels[self, 'load'])
        self.__save .SetLabel(strings.labels[self, 'save'])
        self.__clear.SetLabel(strings.labels[self, 'clear'])
        self.__hint .SetLabel(strings.labels[self, 'hint'])

        self.__btnSizer.Add((10, 1))
        self.__btnSizer.Add(self.__hint, flag=wx.EXPAND, proportion=1)
        self.__btnSizer.Add(self.__load)
        self.__btnSizer.Add((10, 1))
        self.__btnSizer.Add(self.__save)
        self.__btnSizer.Add((10, 1))
        self.__btnSizer.Add(self.__clear)
        self.__btnSizer.Add((10, 1))

        self.__sizer.Add(self.__btnSizer,
                         flag=wx.EXPAND | wx.TOP | wx.BOTTOM,
                         border=3)
        self.__sizer.Add(self.__list,
                         flag=wx.EXPAND,
                         proportion=1)
        self.SetSizer(self.__sizer)
        self.Layout()

        self.__list .Bind(elistbox.EVT_ELB_SELECT_EVENT, self.__onListSelect)
        self.__load .Bind(wx.EVT_BUTTON,                 self.__onLoad)
        self.__save .Bind(wx.EVT_BUTTON,                 self.__onSave)
        self.__clear.Bind(wx.EVT_BUTTON,                 self.__onClear)

        self.__canvas.events.register(self.name,
                                      self.__profileChanged,
                                      'profile')
        self.__profileChanged()


    def destroy(self):
        """Must be called when this ``LocationHistoryPanel`` is no longer
        needed.
        """
        self.__deregisterProfile()
        self.__canvas.events.deregister(self.name, 'profile')
        self.__canvas = None
        fslpanel.FSLeyesPanel.destroy(self)


    def getHistory(self):
        """Returns a list containing the currently displayed location history.
        Each entry in the list is a tuple containing the coordinates, and any
        comment that the user has addded.
        """

        history = []

        for i in range(self.__list.GetCount()):
            location = self.__list.GetItemData(i)
            comment  = self.__list.GetItemLabel(i)
            history.append((location, comment))

        return history


    def __registerProfile(self, profile):
        """Registers a mouse event listener with the given ``profile``. """
        if profile is not None:
            self.__profile = profile
            self.__profile.registerHandler('LeftMouseUp',
                                           self.name,
                                           self.__onMouseUp)


    def __deregisterProfile(self):
        """De-registers listeners that have previously been registered with a
        :class:`.Profile` object.
        """
        if self.__profile is not None:
            self.__profile.deregisterHandler('LeftMouseUp', self.name)
            self.__profile = None


    def __profileChanged(self, *a):
        """Called when the :attr:`.CanvasPanel.profile` changes. Re-registers
        mouse event listeners with the new :class:`.Profile` object.
        """
        self.__deregisterProfile()
        self.__registerProfile(self.__canvas.currentProfile)


    def __addLocation(self, worldLoc, comment=None):
        """Add a location to the location history.

        :arg worldLoc: Location in world coordinates
        :arg comment:  Comment about the location
        """

        if comment is None:
            comment = ''

        label = '[{:7.2f}, {:7.2f}, {:7.2f}]'.format(*worldLoc)
        label = wx.StaticText(self.__list, label=label)

        self.__list.Append(comment, clientData=worldLoc, extraWidget=label)


    def __onMouseUp(self, ev, canvas, mouseLoc, canvasLoc):
        """Called on mouse up events. Adds the mouse location (in the world
        coordinate system) to the location history list.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        opts    = self.displayCtx.getOpts(overlay)
        overlay = opts.referenceImage

        if overlay is not None:
            opts     = self.displayCtx.getOpts(overlay)
            worldLoc = opts.transformCoords(canvasLoc, 'display', 'world')
        else:
            worldLoc = canvasLoc

        self.__addLocation(worldLoc)

        while self.__list.GetCount() > self.__limit:
            self.__list.Delete(0)


    def __onListSelect(self, ev):
        """Called when a location is selected from the location history list.

        Sets the :attr:`.DisplayContext.worldLocation` accordingly.
        """
        self.displayCtx.worldLocation = ev.data


    def __onLoad(self, ev):
        """Called when the *load* button is pushed. Prompts the user to select
        a file, then loads a history from that file.
        """

        msg     = strings.messages[self, 'load']
        fromDir = fslsettings.read('loadSaveOverlayDir')
        dlg     = wx.FileDialog(self,
                                message=msg,
                                defaultDir=fromDir,
                                wildcard='*.txt',
                                style=wx.FD_OPEN)

        if dlg.ShowModal() != wx.ID_OK:
            return

        errTitle = strings.titles[  self, 'loadError']
        errMsg   = strings.messages[self, 'loadError'].format(dlg.GetPath())
        with status.reportIfError(errTitle, errMsg, raiseError=False):
            history = loadLocationHistory(dlg.GetPath())

        self.__list.Clear()
        for loc, comment  in history:
            self.__addLocation(loc, comment)


    def __onSave(self, ev):
        """Called when the *save* button is pushed. Prompts the user to select
        a file name, then saves the current history to that file.
        """
        history = self.getHistory()
        msg     = strings.messages[self, 'save']
        defDir  = fslsettings.read('loadSaveOverlayDir')
        dlg     = wx.FileDialog(self,
                                message=msg,
                                defaultDir=defDir,
                                wildcard='*.txt',
                                style=wx.FD_SAVE)

        if dlg.ShowModal() != wx.ID_OK:
            return

        errTitle = strings.titles[  self, 'saveError']
        errMsg   = strings.messages[self, 'saveError'].format(dlg.GetPath())
        with status.reportIfError(errTitle, errMsg, raiseError=False):
            saveLocationHistory(history, dlg.GetPath())


    def __onClear(self, ev):
        """Called when the *clear* button is pushed. Clears the current
        history.
        """
        self.__list.Clear()


def loadLocationHistory(filename):
    """Loads a location history from the given ``filename``. A location history
    file contains one location on each line, where the X, Y, and Z coordinates
    are separated by a space character. All remaining characters on the line
    are treated as the location comment. For example::

        1.27 4.23 1.63 this is a comment
        5.25 2.66 1.23

    Returns a list containing the location history, in the same format as that
    returned by :meth:`.LocationHistoryPanel.getHistory`.
    """

    history = []

    with open(filename, 'rt') as f:

        lines = f.read().strip().split('\n')

        for line in lines:
            line = line.strip()

            if line == '':
                continue

            try:
                parts = line.split(None, 3)
                loc   = [float(parts[0]),
                         float(parts[1]),
                         float(parts[2])]

                if len(parts) > 3: comment = parts[3]
                else:              comment = ''

                history.append((loc, comment))

            except Exception:
                raise ValueError('Invalid location history '
                                 'format: {}'.format(line))

    return history


def saveLocationHistory(history, filename):
    """Saves the given location ``history`` to the given ``filename``. See
    :func:`loadLocationHistory` and :meth:`LocationHistoryPanel.getHistory`.
    """

    lines = []

    for loc, comment in history:
        loc = '{:0.8f} {:0.8f} {:0.8f}'.format(*loc)
        lines.append('{} {}'.format(loc, comment))

    with open(filename, 'wt') as f:
        f.write('\n'.join(lines))



#!/usr/bin/env python
#
# __init__.py - This package contains a collection of wx.Panel subclasses
#               which provide a view of an image collection.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``views`` package is the home of all *FSLeyes views*, as described in
the :mod:`~fsleyes` package documentation.  It contains a collection of
:class:`FSLeyesPanel` sub-classes which provide a view of the overlays in an
:class:`.OverlayList`.  The :class:`.ViewPanel` is the base-class for all
views.
"""



#!/usr/bin/env python
#
# maskopts.py - The MaskOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`MaskOpts` class, which defines settings
for displaying an :class:`.Image` overlay as a binary mask.
"""


import copy
import logging

import fsleyes_props   as props
import fsleyes.gl      as fslgl
from . import             volumeopts
from . import             niftiopts


log = logging.getLogger(__name__)


class MaskOpts(niftiopts.NiftiOpts):
    """The ``MaskOpts`` class defines settings for displaying an
    :class:`.Image` overlay as a binary mask.
    """


    threshold = props.Bounds(ndims=1)
    """The mask threshold range - values outside of this range are not
    displayed.
    """


    invert = props.Boolean(default=False)
    """If ``True``, the :attr:`threshold` range is inverted - values
    inside the range are not shown, and values outside of the range are shown.
    """


    colour = props.Colour()
    """The mask colour."""


    outline = props.Boolean(default=False)
    """If ``True`` only the outline of the mask will be shown.  If ``False``,
    the filled mask will be displayed.
    """


    outlineWidth = props.Int(minval=0, maxval=10, default=1, clamped=True)
    """Width of mask outline, if :attr:``outline` is ``True``.  This value is
    in terms of pixels.
    """


    interpolation = copy.copy(volumeopts.VolumeOpts.interpolation)


    def __init__(self, overlay, *args, **kwargs):
        """Create a ``MaskOpts`` instance for the given overlay. All arguments
        are passed through to the :class:`.NiftiOpts` constructor.
        """

        # We need GL >= 2.1 for
        # spline interpolation
        if float(fslgl.GL_COMPATIBILITY) < 2.1:
            interp = self.getProp('interpolation')
            interp.removeChoice('spline', instance=self)
            interp.updateChoice('linear', instance=self, newAlt=['spline'])

        kwargs['nounbind'] = ['interpolation']

        # Initialise threshold from data reange. Do
        # this before __init__, in case we need to
        # inherit settings from the parent instance
        dmin, dmax = overlay.dataRange
        dlen       = dmax - dmin
        doff       = dlen / 100.0

        self.threshold.xmin = dmin - doff
        self.threshold.xmax = dmax + doff
        self.threshold.xlo  = dmin + doff
        self.threshold.xhi  = dmax + doff

        niftiopts.NiftiOpts.__init__(self, overlay, *args, **kwargs)

        overlay.register(self.name,
                         self.__dataRangeChanged,
                         topic='dataRange',
                         runOnIdle=True)

        # The master MaskOpts instance makes
        # sure that colour[3] and Display.alpha
        # are consistent w.r.t. each other.
        self.__registered = self.getParent() is None
        if self.__registered:
            self.display.addListener('alpha',
                                     self.name,
                                     self.__alphaChanged,
                                     immediate=True)
            self        .addListener('colour',
                                     self.name,
                                     self.__colourChanged,
                                     immediate=True)


    def destroy(self):
        """Removes some property listeners and calls
        :meth:`.NitfiOpts.destroy`.
        """

        self.overlay.deregister(self.name, topic='dataRange')

        if self.__registered:
            self.display.removeListener('alpha',  self.name)
            self        .removeListener('colour', self.name)

        niftiopts.NiftiOpts.destroy(self)


    def __dataRangeChanged(self, *a):
        """Called when the :attr:`~fsl.data.image.Image.dataRange` changes.
        Updates the :attr:`threshold` limits.
        """
        dmin, dmax   = self.overlay.dataRange
        dRangeLen    = abs(dmax - dmin)
        dminDistance = dRangeLen / 100.0

        self.threshold.xmin = dmin - dminDistance
        self.threshold.xmax = dmax + dminDistance

        # If the threshold was
        # previously unset, grow it
        if self.threshold.x == (0, 0):
            self.threshold.x = (0, dmax + dminDistance)


    def __colourChanged(self, *a):
        """Called when :attr:`.colour` changes. Updates :attr:`.Display.alpha`
        from the alpha component.
        """

        alpha = self.colour[3] * 100

        log.debug('Propagating MaskOpts.colour[3] to '
                  'Display.alpha [{}]'.format(alpha))

        with props.skip(self.display, 'alpha', self.name):
            self.display.alpha = alpha


    def __alphaChanged(self, *a):
        """Called when :attr:`.Display.alpha` changes. Updates the alpha
        component of :attr:`.colour`.
        """

        alpha       = self.display.alpha / 100.0
        r, g, b, _  = self.colour

        log.debug('Propagating Display.alpha to MaskOpts.'
                  'colour[3] [{}]'.format(alpha))

        with props.skip(self, 'colour', self.name):
            self.colour = r, g, b, alpha



#!/usr/bin/env python
#
# wxgllightboxcanvas.py - THe  WXGLLightBoxCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`WXGLSliceCanvas` class, which is a
:class:`.SliceCanvas` for use in a :mod:`wx` application.
"""
"""This module provides the :class:`WXGLLightBoxCanvas`, which is a
:class:`.LightBoxCanvas` for use in a :mod:`wx` application.
"""


import wx
import wx.siplib   as sip
import wx.glcanvas as wxgl

import fsleyes.gl  as fslgl
from . import         lightboxcanvas


class WXGLLightBoxCanvas(lightboxcanvas.LightBoxCanvas,
                         fslgl.WXGLCanvasTarget,
                         wxgl.GLCanvas):
    """The ``WXGLLightBoxCanvas`` is a :class:`.LightBoxCanvas`, a
    :class:`wx.glcanvas.GLCanvas` and a :class:`.WXGLCanvasTarget`. If you
    want to use a :class:`.LightBoxCanvas` in your :mod:`wx` application,
    then you should use a ``WXGLLightBoxCanvas``.

    .. note:: The ``WXGLLightBoxCanvas`` assumes the existence of the
              :meth:`.LightBoxCanvas._updateDisplayBounds` method.
    """

    __metaclass__ = sip.wrappertype

    def __init__(self, parent, overlayList, displayCtx, zax=0):
        """Create a ``WXGLLightBoxCanvas``. See
        :meth:`.LightBoxCanvas.__init__` for details on the arguments.
        """

        attrs = fslgl.WXGLCanvasTarget.displayAttribues()

        wxgl.GLCanvas                .__init__(self, parent, **attrs)
        fslgl.WXGLCanvasTarget       .__init__(self)
        lightboxcanvas.LightBoxCanvas.__init__(self,
                                               overlayList,
                                               displayCtx,
                                               zax)

        # When the canvas is resized, we have to update
        # the display bounds to preserve the aspect ratio
        def onResize(ev):
            self._updateDisplayBounds()
            ev.Skip()
        self.Bind(wx.EVT_SIZE, onResize)


    def destroy(self):
        """Must be called when this ``WXGLLightBoxCanvas`` is no longer
        needed. Clears some event listeners and calls the base class
        ``destroy`` method.
        """
        self.Unbind(wx.EVT_SIZE)
        lightboxcanvas.LightBoxCanvas.destroy(self)
        fslgl.WXGLCanvasTarget       .destroy(self)



#!/usr/bin/env python
#
# __init__.py - ARB vertex/fragment program management.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``arbp`` package is the home of the :class:`.ARBPShader` class, a class
which may be used to manage OpenGL ``ARB_vertex_program`` and
``ARB_fragment_program`` shader programs.  The :class:`.ARBPShader` class is
defined in the :mod:`.program` module, and uses some functions defined in the
:mod:`.parse` module.

See also the :mod:`.glsl` package.
"""



#!/usr/bin/env python
#
# glmask.py - The GLMask class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLMask` class, which implements
functionality for rendering an :class:`.Image` overlay as a binary mask.
"""

import logging

import OpenGL.GL                 as gl

import fsl.utils.idle            as idle
import fsleyes.colourmaps        as colourmaps
import fsleyes.gl                as fslgl
import fsleyes.gl.textures       as textures
import fsleyes.gl.routines       as glroutines
import fsleyes.gl.resources      as glresources
import fsleyes.gl.shaders.filter as glfilter
from . import                       glimageobject


log = logging.getLogger(__name__)


class GLMask(glimageobject.GLImageObject):
    """The ``GLMask`` class encapsulates logic to render an :class:`.Image`
    instance as a binary mask in OpenGL.

    When created, a ``GLMask`` instance assumes that the provided
    :class:`.Image` instance has a :attr:`.Display.overlayType` of ``mask``,
    and that its associated :class:`.Display` instance contains a
    :class:`.MaskOpts` instance, containing mask-specific display properties.


    **Textures**

    A ``GLMask`` will use up to two textures:

      - An :class:`.ImageTexture` for storing the 3D image data. This texture
        is bound to texture unit 0.

      - A :class:`.RenderTexture`, used for edge filtering if necessary. This
        texture will be bound to texture unit 1.


    **2D rendering**


    On 2D canvases, A ``GLMask`` is rendered similarly to a :class:`.GLVolume`
    - a 2D slice is taken through the 3D image texture. If the
    :attr:`.MaskOpts.outline` property is active, this slice is rendered to
    an off-screen texture, which is then passed through an edge filter (see
    the :mod:`.filters` module).


    **Version dependent modules**


    The ``GLMask`` class makes use of the functions defined in the
    :mod:`.gl14.glmask_funcs` or the :mod:`.gl21.glmask_funcs` modules,
    which provide OpenGL version specific details for rendering.


    These version dependent modules must provide the following functions:

    ============================= =====================================
    ``init(GLMask)``              Perform any necessary initialisation.
    ``destroy(GLMask)``           Perform any necessary cleanup
    ``compileShaders(GLMask)``    (Re-)compile the shader program
    ``updateShaderState(GLMask)`` Update the shader program state
    ``draw2D(GLMask, ...)``       Draw a slice of the image
    ``drawAll(GLMask, ...)``      Draw multiple slices of the image
    ============================= =====================================
    """


    def __init__(self, image, overlayList, displayCtx, canvas, threedee):
        """Create a ``GLMask``.

        :arg image:       The :class:`.Image` instance.
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
        :arg canvas:      The canvas doing the drawing.
        :arg threedee:    2D or 3D rendering
        """
        glimageobject.GLImageObject.__init__(self,
                                             image,
                                             overlayList,
                                             displayCtx,
                                             canvas,
                                             threedee)

        # The shader attribute will be created
        # by the glmask_funcs module
        self.shader        = None
        self.imageTexture  = None
        self.edgeFilter    = glfilter.Filter('edge', texture=1)
        self.renderTexture = textures.RenderTexture(
            self.name, interp=gl.GL_LINEAR, rttype='c')

        self.addDisplayListeners()
        self.refreshImageTexture()

        def init():
            fslgl.glmask_funcs.init(self)
            self.notify()

        idle.idleWhen(init, self.textureReady)


    def destroy(self):
        """Must be called when this ``GLMask`` is no longer needed. Destroys
        the :class:`.ImageTexture`.
        """
        self.edgeFilter.destroy()
        self.renderTexture.destroy()
        self.imageTexture.deregister(self.name)
        glresources.delete(self.imageTexture.name)

        self.removeDisplayListeners()
        fslgl.glmask_funcs.destroy(self)
        glimageobject.GLImageObject.destroy(self)


    def ready(self):
        """Returns ``True`` if this ``GLMask`` is ready to be drawn, ``False``
        otherwise.
        """
        return self.shader is not None and self.textureReady()


    def textureReady(self):
        """Returns ``True`` if the ``imageTexture`` is ready to be used,
        ``False`` otherwise.
        """
        return self.imageTexture is not None and self.imageTexture.ready()


    def updateShaderState(self, *args, **kwargs):
        """Calls :func:`.gl14.gllabel_funcs.updateShaderState` or
        :func:`.gl21.gllabel_funcs.updateShaderState`, and
        :meth:`.Notifier.notify`. Uses :func:`.idle.idleWhen` to ensure that
        they don't get called until :meth:`ready` returns ``True``.
        """
        alwaysNotify = kwargs.pop('alwaysNotify', None)

        def func():
            if fslgl.glmask_funcs.updateShaderState(self) or alwaysNotify:
                self.notify()

        idle.idleWhen(func,
                      self.ready,
                      name=self.name,
                      skipIfQueued=True)


    def addDisplayListeners(self):
        """Adds a bunch of listeners to the :class:`.Display` object, and the
        associated :class:`.MaskOpts` instance, which define how the mask
        image should be displayed.
        """

        display = self.display
        opts    = self.opts
        name    = self.name

        def update(*a):
            self.updateShaderState(alwaysNotify=True)

        display.addListener('alpha',         name, update, weak=False)
        display.addListener('brightness',    name, update, weak=False)
        display.addListener('contrast',      name, update, weak=False)
        opts   .addListener('colour',        name, update, weak=False)
        opts   .addListener('threshold',     name, update, weak=False)
        opts   .addListener('invert',        name, update, weak=False)
        opts   .addListener('outlineWidth',  name, update, weak=False)
        opts   .addListener('outline',       name, self.notify)
        opts   .addListener('transform',     name, self.notify)
        opts   .addListener('volume',        name, self.__volumeChanged)
        opts   .addListener('interpolation', name, self.__interpChanged)

        # See comment in GLVolume.addDisplayListeners about this
        self.__syncListenersRegistered = opts.getParent() is not None

        if self.__syncListenersRegistered:
            opts.addSyncChangeListener(
                'volume', name, self.refreshImageTexture)


    def removeDisplayListeners(self):
        """Removes all the listeners added by :meth:`addDisplayListeners`. """

        display = self.display
        opts    = self.opts
        name    = self.name

        display.removeListener('alpha',         name)
        display.removeListener('brightness',    name)
        display.removeListener('contrast',      name)
        opts   .removeListener('colour',        name)
        opts   .removeListener('threshold',     name)
        opts   .removeListener('invert',        name)
        opts   .removeListener('outline',       name)
        opts   .removeListener('outlineWidth',  name)
        opts   .removeListener('transform',     name)
        opts   .removeListener('volume',        name)
        opts   .removeListener('interpolation', name)

        if self.__syncListenersRegistered:
            opts.removeSyncChangeListener('volume', name)


    def refreshImageTexture(self):
        """Makes sure that the :class:`.ImageTexture`, used to store the
        :class:`.Image` data, is up to date.
        """

        opts     = self.opts
        texName  = '{}_{}' .format(type(self).__name__, id(self.image))
        unsynced = (opts.getParent() is None or
                    not opts.isSyncedToParent('volume'))

        if unsynced:
            texName = '{}_unsync_{}'.format(texName, id(opts))

        if self.imageTexture is not None:
            if self.imageTexture.name == texName:
                return

            self.imageTexture.deregister(self.name)
            glresources.delete(self.imageTexture.name)

        if opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                            interp = gl.GL_LINEAR

        self.imageTexture = glresources.get(
            texName,
            textures.ImageTexture,
            texName,
            self.image,
            interp=interp,
            volume=opts.index()[3:],
            notify=False)

        self.imageTexture.register(self.name, self.__imageTextureChanged)


    def getColour(self):
        """Prepares and returns the mask colour for use in the fragment shader.
        """
        display = self.display
        opts    = self.opts
        alpha   = display.alpha / 100.0
        colour  = opts.colour

        colour = colour[:3]
        colour = colourmaps.applyBricon(colour,
                                        display.brightness / 100.0,
                                        display.contrast   / 100.0)

        return list(colour) + [alpha]


    def getThreshold(self):
        """Prepares and returns the mask thresholds for use in the fragment
        shader.
        """
        opts  = self.opts
        xform = self.imageTexture.invVoxValXform
        lo    = opts.threshold[0] * xform[0, 0] + xform[0, 3]
        hi    = opts.threshold[1] * xform[0, 0] + xform[0, 3]
        return (lo, hi)


    def preDraw(self, xform=None, bbox=None):
        """Binds the :class:`.ImageTexture` and calls the version-dependent
        ``preDraw`` function.
        """

        w, h = self.canvas.GetSize()
        rtex = self.renderTexture

        rtex.shape = w, h
        with rtex.target():
            gl.glClearColor(0, 0, 0, 0)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        self.imageTexture.bindTexture(gl.GL_TEXTURE0)


    def draw2D(self, zpos, axes, xform=None, bbox=None):
        """Calls the version-dependent ``draw2D`` function, then applies
        the edge filter if necessary.
        """

        opts = self.opts

        if not opts.outline:
            fslgl.glmask_funcs.draw2D(self, zpos, axes, xform, bbox)
            return

        owidth     = float(opts.outlineWidth)
        rtex       = self.renderTexture
        w, h       = self.canvas.GetSize()
        lo, hi     = self.canvas.getViewport()
        xax        = axes[0]
        yax        = axes[1]
        xmin, xmax = lo[xax], hi[xax]
        ymin, ymax = lo[yax], hi[yax]
        offsets    = [owidth / w, owidth / h]

        # Draw the mask to the off-screen texture
        with glroutines.disabled(gl.GL_BLEND), rtex.target(xax, yax, lo, hi):
            fslgl.glmask_funcs.draw2D(self, zpos, axes, xform, bbox)

        # Run the texture through an edge detection
        # filter, drawing the result to screen
        self.edgeFilter.set(offsets=offsets, outline=1)
        self.edgeFilter.apply(
            rtex, zpos, xmin, xmax, ymin, ymax, xax, yax,
            textureUnit=gl.GL_TEXTURE1)


    def drawAll(self, axes, zposes, xforms):
        """Calls the version-dependent ``drawAll`` function, then applies
        the edge filter if necessary.
        """

        opts = self.opts
        rtex = self.renderTexture

        # Is taking max(z) hacky? It seems to work ok.
        zpos       = max(zposes)
        owidth     = opts.outlineWidth
        w, h       = self.canvas.GetSize()
        lo, hi     = self.canvas.getViewport()
        xax        = axes[0]
        yax        = axes[1]
        xmin, xmax = lo[xax], hi[xax]
        ymin, ymax = lo[yax], hi[yax]
        offsets    = [owidth / w, owidth / h]

        # Draw all slices to the off-screen texture
        with glroutines.disabled(gl.GL_BLEND), rtex.target(xax, yax, lo, hi):
            fslgl.glmask_funcs.drawAll(self, axes, zposes, xforms)

        # if no outline, draw the texture directly
        if not opts.outline:
            rtex.drawOnBounds(
                zpos, xmin, xmax, ymin, ymax, xax, yax,
                textureUnit=gl.GL_TEXTURE1)

        else:
            # Run the texture through an edge detection
            # filter, drawing the result to screen
            self.edgeFilter.set(offsets=offsets, outline=1)
            self.edgeFilter.apply(
                rtex, zpos, xmin, xmax, ymin, ymax, xax, yax,
                textureUnit=gl.GL_TEXTURE1)


    def draw3D(self, *args, **kwargs):
        """Does nothing. """
        pass


    def postDraw(self, xform=None, bbox=None):
        """Unbinds the ``ImageTexture``. """
        self.imageTexture.unbindTexture()


    def __volumeChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.volume` changes. Updates the
        image texture.
        """
        self.imageTexture.set(volume=self.opts.index()[3:])


    def __interpChanged(self, *a):
        """Called when the :attr:`.MaskOpts.interpolation` changes. Updates the
        image texture.
        """
        if self.opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                                 interp = gl.GL_LINEAR
        self.imageTexture.set(interp=interp)


    def __imageTextureChanged(self, *a):
        """Called when the image texture data has changed. Triggers a refresh.
        """
        self.updateShaderState(alwaysNotify=True)


    def __imageSyncChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.volume` property is synchronised
        or un-synchronised. Calls :meth:`refreshImageTexture` and
        :meth:`updateShaderState`.
        """
        self.refreshImageTexture()
        self.updateShaderState(alwaysNotify=True)



#!/usr/bin/env python
#
# glmask_funcs.py - OpenGL 1.4 functions used by the GLMask class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLMask`
class to render :class:`.Image` overlays in an OpenGL 1.4 compatible manner.
"""


import fsleyes.gl.shaders as shaders
from . import                glvolume_funcs


def init(self):
    """Calls the :func:`compileShaders` and :func:`updateShaderState`
    functions.
    """
    self.shader = None
    compileShaders(   self)
    updateShaderState(self)


def destroy(self):
    """Destroys the shader programs. """
    self.shader.destroy()
    self.shader = None


def compileShaders(self):
    """Loads the vertex/fragment shader source code, and creates a
    :class:`.ARBPShader` program.
    """

    if self.shader is not None:
        self.shader.destroy()

    vertSrc  = shaders.getVertexShader(  'glvolume')
    fragSrc  = shaders.getFragmentShader('glmask')
    textures = {
        'imageTexture' : 0,
    }

    self.shader = shaders.ARBPShader(vertSrc,
                                     fragSrc,
                                     shaders.getShaderDir(),
                                     textures)


def updateShaderState(self):
    """Updates all shader program variables. """

    if not self.ready():
        return

    opts       = self.opts
    shader     = self.shader
    colour     = self.getColour()
    threshold  = list(self.getThreshold())

    if opts.invert: threshold += [ 1, 0]
    else:           threshold += [-1, 0]

    shader.load()
    shader.setFragParam('threshold', threshold)
    shader.setFragParam('colour',    colour)
    shader.unload()

    return True


def draw2D(self, zpos, axes, xform=None, bbox=None):
    """Draws a 2D slice at the given ``zpos``. Uses the
    :func:`.gl14.glvolume_funcs.draw2D` function.
    """

    self.shader.load()
    self.shader.loadAtts()
    glvolume_funcs.draw2D(self, zpos, axes, xform, bbox)
    self.shader.unloadAtts()
    self.shader.unload()


def drawAll(self, axes, zposes, xforms):
    """Draws all specified slices. Uses the
    :func:`.gl14.glvolume_funcs.drawAll` function.
    """
    self.shader.load()
    self.shader.loadAtts()
    glvolume_funcs.drawAll(self, axes, zposes, xforms)
    self.shader.unloadAtts()
    self.shader.unload()



#!/usr/bin/env python
#
# glcomplex.py - The GLComplex class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLComplex` class, for displaying
:class:`.Image` overlays with a complex data type.
"""


from . import glvolume


class GLComplex(glvolume.GLVolume):
    """The ``GLComplex`` class is a sub-class of :class:`.GLVolume`, specialised
    for displaying :class:`.Image` overlays with a complex data type.


    The only additional behaviour this class provides is refreshing the
    :class:`.ImageTexture` data whenever the :attr:`.ComplexOpts.component`
    property changes.
    """


    def addDisplayListeners(self):
        """Overrides :meth:`VolumeOpts.addDisplayListeners`. Calls that
        method, and also adds additional listeners.
        """
        self.opts.addListener('component', self.name, self.__componentChanged)
        return glvolume.GLVolume.addDisplayListeners(self)


    def removeDisplayListeners(self):
        """Overrides :meth:`VolumeOpts.removeDisplayListeners`. Calls that
        method, and also removes additional listeners.
        """
        self.opts.removeListener('component', self.name)
        return glvolume.GLVolume.removeDisplayListeners(self)


    def refreshImageTexture(self):
        """Overrides :meth:`.GLVolume.refreshImageTexture`. Calls that
        method, passing it a prefilter function to extract the complex
        component from the image data.
        """
        pfunc  = self.getPrefilterFunc()
        prfunc = self.getPrefilterRangeFunc()
        glvolume.GLVolume.refreshImageTexture(self,
                                              prefilter=pfunc,
                                              prefilterRange=prfunc)


    def getPrefilterFunc(self):
        """Returns a function which extracts the component to be displayed
        from the image data. Used as the prefilter function by the
        :class:`.ImageTexture`

        See the :attr:`ComplexOpts.component` property.
        """

        opts      = self.opts
        component = opts.component

        if   component == 'real':  return opts.getReal
        elif component == 'imag':  return opts.getImaginary
        elif component == 'mag':   return opts.getMagnitude
        elif component == 'phase': return opts.getPhase


    def getPrefilterRangeFunc(self):
        """Returns a function which returns the minimum/maximum of the
        current component. Used as the prefilterRange function by the
        :class:`.ImageTexture`.
        """
        def pr(*a):
            return self.opts.getDataRange()
        return pr


    def __componentChanged(self, *a):
        """Called when the :attr:`component` changes. Updates the image texture
        data.
        """
        # We only want the image texture data
        # to be updated once, despite multiple
        # calls to set() (e.g. from three
        # GLComplex objects in an ortho panel).
        #
        # The ComplexOpts class has static
        # methods for obtaining the real/imag/
        # mag/phase components from the data.
        # We can use these as the texture
        # prefilter function - the image
        # texture will only refresh its data
        # when the prefilter function changes,
        # which will only be on the first call
        # (the prefilter functions are static,
        # so subsequent calls will pass in the
        # same function object)
        pfunc  = self.getPrefilterFunc()
        prfunc = self.getPrefilterRangeFunc()
        self.imageTexture.set(prefilter=pfunc,
                              prefilterRange=prfunc,
                              volRefresh=False)



#!/usr/bin/env python
#
# __init__.py - Functions for OpenGL 2.1 rendering.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This package contains modules for rendering various :class:`.GLObject`
types in an OpenGL 2.1 compatible manner.
"""


from . import glvolume_funcs       # noqa
from . import glrgbvolume_funcs    # noqa
from . import glrgbvector_funcs    # noqa
from . import gllinevector_funcs   # noqa
from . import glmesh_funcs         # noqa
from . import glmask_funcs         # noqa
from . import gllabel_funcs        # noqa
from . import gltensor_funcs       # noqa
from . import glsh_funcs           # noqa
from . import glmip_funcs          # noqa



#!/usr/bin/env python
#
# timeseriespanel.py - The TimeSeriesPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TimeSeriesPanel`, which is a *FSLeyes
view* for displaying time series data from :class:`.Image` overlays.
"""


import logging
import itertools as it

import numpy as np

import fsl.data.featimage                 as fslfeatimage
import fsl.data.melodicimage              as fslmelimage
import fsl.data.image                     as fslimage
import fsl.data.mesh                      as fslmesh
import fsleyes_props                      as props

import fsleyes.overlay                    as fsloverlay
import fsleyes.strings                    as strings
import fsleyes.profiles.timeseriesprofile as timeseriesprofile
import fsleyes.plotting.timeseries        as timeseries
import fsleyes.views.plotpanel            as plotpanel


log = logging.getLogger(__name__)


class TimeSeriesPanel(plotpanel.OverlayPlotPanel):
    """The ``TimeSeriesPanel`` is an :class:`.OverlayPlotPanel` which plots
    time series data from overlays. A ``TimeSeriesPanel`` looks something like
    the following:


    .. image:: images/timeseriespanel.png
       :scale: 50%
       :align: center


    A ``TimeSeriesPanel`` plots one or more :class:`.DataSeries` instances,
    which encapsulate time series data from an overlay. All time series
    related ``DataSeries`` classes are defined in the
    :mod:`.plotting.timeseries` module; these are all sub-classes of the
    :class:`.DataSeries` class - see the :class:`.PlotPanel` and
    :class:`.OverlayPlotPanel` documentation for more details:

    .. autosummary::
       :nosignatures:

       .timeseries.VoxelTimeSeries
       .timeseries.ComplexTimeSeries
       .timeseries.FEATTimeSeries
       .timeseries.MelodicTimeSeries
       .timeseries.MeshTimeSeries


    **Control panels**


    Some *FSLeyes control* panels are associated with the
    :class:`.TimeSeriesPanel`, and can be added/removed via
    :meth:`.ViewPanel.togglePanel`.

    .. autosummary::
       :nosignatures:

       ~fsleyes.controls.plotlistpanel.PlotListPanel
       ~fsleyes.controls.timeseriescontrolpanel.TimeSeriesControlPanel


    **FEATures**


    The ``TimeSeriesPanel`` has some extra functionality for
    :class:`.FEATImage` overlays. For these overlays, a
    :class:`.FEATTimeSeries` instance is plotted, instead of a regular
    :class:`.TimeSeries` instance. The ``FEATTimeSeries`` class, in turn, has
    the ability to generate more ``TimeSeries`` instances which represent
    various aspects of the FEAT model fit. See the :class:`.FEATTimeSeries`
    and the :class:`.TimeSeriesControlPanel` classes for more details.


    **Melodic features**


    The ``TimeSeriesPanel`` also has some functionality for
    :class:`.MelodicImage` overlays - a :class:`.MelodicTimeSeries` instance
    is used to plot the component time courses for the current component (as
    defined by the :attr:`.NiftiOpts.volume` property).
    """


    usePixdim = props.Boolean(default=False)
    """If ``True``, the X axis data is scaled by the pixdim value of the
    selected overlay (which, for FMRI time series data is typically set
    to the TR time).
    """


    plotMode = props.Choice(('normal', 'demean', 'normalise', 'percentChange'))
    """Options to scale/offset the plotted time courses.

    ================= =======================================================
    ``normal``        The data is plotted with no modifications
    ``demean``        The data is demeaned (i.e. plotted with a mean of 0)
    ``normalise``     The data is normalised to lie in the range ``[-1, 1]``.
    ``percentChange`` The data is scaled to percent changed
    ================= =======================================================
    """


    plotMelodicICs = props.Boolean(default=True)
    """If ``True``, the component time courses are plotted for
    :class:`.MelodicImage` overlays (using a :class:`.MelodicTimeSeries`
    instance). Otherwise, ``MelodicImage`` overlays are treated as regular
    4D :class:`.Image` overlays (a :class:`.VoxelTimeSeries` instance is used).
    """


    @staticmethod
    def defaultLayout():
        """Returns a list of control panel types to be added for the default
        time series panel layout.
        """
        return ['TimeSeriesToolBar',
                'OverlayListPanel',
                'PlotListPanel']


    @staticmethod
    def controlOrder():
        """Returns a list of control panel names, specifying the order in
        which they should appear in the  FSLeyes ortho panel settings menu.
        """
        return ['OverlayListPanel',
                'PlotListPanel',
                'TimeSeriesToolBar',
                'TimeSeriesControlPanel']


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``TimeSeriesPanel``.

        :arg parent:      A :mod:`wx` parent object.
        :arg overlayList: An :class:`.OverlayList` instance.
        :arg displayCtx:  A :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """

        # If the currently selected image is from
        # a FEAT analysis, and the corresponding
        # filtered_func_data is loaded, enable the
        # initial state of the time course for
        # that filtered_func_data
        featImage = fsloverlay.findFEATImage(
            overlayList,
            displayCtx.getSelectedOverlay())

        if featImage is None: initialState = None
        else:                 initialState = {featImage : True}

        plotpanel.OverlayPlotPanel.__init__(
            self,
            parent,
            overlayList,
            displayCtx,
            frame,
            initialState=initialState)

        self.addListener('plotMode',  self.name, self.draw)
        self.addListener('usePixdim', self.name, self.draw)
        self.addListener('plotMelodicICs',
                         self.name,
                         self.__plotMelodicICsChanged)

        self.initProfile(timeseriesprofile.TimeSeriesProfile)


    def destroy(self):
        """Removes some listeners, and calls the :meth:`.PlotPanel.destroy`
        method.
        """

        self.removeListener('plotMode',       self.name)
        self.removeListener('usePixdim',      self.name)
        self.removeListener('plotMelodicICs', self.name)

        plotpanel.OverlayPlotPanel.destroy(self)


    def getActions(self):
        """Overrides :meth:`.ActionProvider.getActions`. Returns all of the
        :mod:`.actions` that are defined on this ``TimeSeriesPanel``.
        """
        actionz = [self.screenshot,
                   self.importDataSeries,
                   self.exportDataSeries]

        names = [a.actionName if a is not None else None for a in actionz]
        return list(zip(names, actionz))


    def draw(self, *a):
        """Overrides :meth:`.PlotPanel.draw`. Passes some :class:`.DataSeries`
        instances to the :meth:`.PlotCanvas.drawDataSeries` method, then
        calls :meth:`.PlotCanvas.drawArtists`.
        """

        if not self or self.destroyed:
            return

        tss = self.getDataSeriesToPlot()
        for ts in tss:

            # Changing the label might trigger
            # another call to this method, as
            # the PlotPanel might have a listener
            # registered on it. Hence the suppress
            with props.suppress(ts, 'label'):
                ts.label = ts.makeLabel()

        xlabel, ylabel = self.__generateDefaultLabels(tss)

        self.canvas.drawDataSeries(extraSeries=tss,
                                   xlabel=xlabel,
                                   ylabel=ylabel)
        self.canvas.drawArtists()


    def createDataSeries(self, overlay):
        """Overrides :meth:`.OverlayPlotPanel.createDataSeries`. Creates and
        returns a :class:`.TimeSeries` instance (or an instance of one of the
        :class:`.TimeSeries` sub-classes) for the specified overlay.

        Returns a tuple containing the following:

          - A :class:`.DataSeries` instance for the given overlay

          - A list of *targets* - objects which have properties that
            influence the state of the ``TimeSeries`` instance.

          - A list of *property names*, one for each target.

        If the given overlay is not compatible (i.e. it has no time series
        data to be plotted), a tuple of ``None`` values is returned.
        """

        displayCtx  = self.displayCtx
        overlayList = self.overlayList
        tsargs      = (overlay, overlayList, displayCtx, self.canvas)

        # Is this a mesh?
        if isinstance(overlay, fslmesh.Mesh):
            ts        = timeseries.MeshTimeSeries(*tsargs)
            opts      = displayCtx.getOpts(overlay)
            targets   = [displayCtx, opts]
            propNames = ['location', 'vertexData']

        # Is this a FEAT filtered_func_data image?
        elif isinstance(overlay, fslfeatimage.FEATImage):

            # If the filtered_func for this FEAT analysis
            # has been loaded, we show its time series.
            ts        = timeseries.FEATTimeSeries(*tsargs)
            targets   = [displayCtx]
            propNames = ['location']

        # If this is a melodic IC image, and we are
        # currently configured to plot component ICs,
        # we use a MelodicTimeSeries object.
        elif isinstance(overlay, fslmelimage.MelodicImage) and \
             self.plotMelodicICs:
            ts        = timeseries.MelodicTimeSeries(*tsargs)
            targets   = [displayCtx.getOpts(overlay)]
            propNames = ['volume']

        # Otherwise it's a normal image
        elif isinstance(overlay, fslimage.Image) and overlay.ndim > 3:

            # Is it a complex data image?
            if overlay.iscomplex:
                ts = timeseries.ComplexTimeSeries(*tsargs)

            # Or just a bog-standard 4D image?
            else:
                ts = timeseries.VoxelTimeSeries(*tsargs)

            # listen to volumeDim for
            # images with >4 dimensions
            opts      = displayCtx.getOpts(overlay)
            targets   = [displayCtx, opts]
            propNames = ['location', 'volumeDim']
        else:
            return None, None, None

        ts.colour    = self.getOverlayPlotColour(overlay)
        ts.lineStyle = self.getOverlayPlotStyle(overlay)
        ts.lineWidth = 2
        ts.alpha     = 1

        return ts, targets, propNames


    def prepareDataSeries(self, ts):
        """Overrides :class:`.PlotPanel.prepareDataSeries`. Given a
        :class:`.DataSeries` instance, scales and normalises the x and y data
        according to the current values of the :attr:`usePixdim` and
        :attr:`plotMode` properties.
        """

        xdata, ydata = ts.getData()

        if (xdata is None) or (ydata is None) or (len(xdata) == 0):
            return xdata, ydata

        if self.usePixdim and isinstance(ts.overlay, fslimage.Image):
            if isinstance(ts.overlay, fslmelimage.MelodicImage):
                xdata = xdata * ts.overlay.tr
            else:
                xdata = xdata * ts.overlay.pixdim[3]

            toffset = ts.overlay.nibImage.header.get('toffset', 0)
            xdata  += toffset

        if self.plotMode == 'demean':
            ydata = ydata - ydata.mean()

        elif self.plotMode == 'normalise':
            ymin  = ydata.min()
            ymax  = ydata.max()
            if not np.isclose(ymin, ymax):
                ydata = 2 * (ydata - ymin) / (ymax - ymin) - 1
            else:
                ydata = np.zeros(len(ydata))

        elif self.plotMode == 'percentChange':
            mean  = ydata.mean()
            ydata = 100 * (ydata / mean) - 100

        return xdata, ydata


    def __plotMelodicICsChanged(self, *a):
        """Called when the :attr:`plotMelodicICs` property changes. Re-creates
        the internally cached :class:`.TimeSeries` instances for all
        :class:`.MelodicImage` overlays in the :class:`.OverlayList`.
        """

        for overlay in self.overlayList:
            if isinstance(overlay, fslmelimage.MelodicImage):
                self.clearDataSeries(overlay)

        self.updateDataSeries()
        self.draw()


    def __generateDefaultLabels(self, timeSeries):
        """Called by :meth:`draw`. If the :attr:`.PlotCanvas.xlabel` or
        :attr:`.PlotCanvas.ylabel` properties are unset, an attempt is made
        to generate default labels.
        """

        xlabel = self.canvas.xlabel
        ylabel = self.canvas.ylabel

        if xlabel is not None:
            return xlabel, ylabel

        if not self.usePixdim:
            xlabel = strings.nifti['t_unit', -1]
            return xlabel, ylabel

        # If all of the overlays related to the data series being
        # plotted:
        #
        #   - are Images
        #   - have the same time unit (as specified in the nifti header)
        #
        # Then a default label is specified
        #
        # n.b. this is not foolproof, as many
        # non-time 4D images will still set
        # the time units to seconds.
        #
        #
        # TODO Non-Image overlays with associated
        #      time series (e.g. MeshOpts)

        # Get all the unique overlays
        overlays = [ts.overlay
                    for ts in it.chain(timeSeries, self.canvas.dataSeries)]
        overlays = set(overlays)

        if not all([isinstance(o, fslimage.Image) for o in overlays]):
            return xlabel, ylabel

        # And all of their time units
        units = [o.timeUnits for o in overlays]

        if len(set(units)) == 1:
            xlabel = strings.nifti.get(('t_unit', units[0]),
                                       'INVALID TIME UNITS')

        return xlabel, ylabel
